"""bpack support for type annotations."""

import re
from typing import NamedTuple, Optional, Type, Union

try:                                                        # pragma: no cover
    from typing_extensions import _tp_cache                 # noqa
    # @COMPATIBILITY: with Python < 3.9
    from typing_extensions import Annotated
    # @COMPATIBILITY: with Python < 3.7 (and Python < 3.8.3)
    from typing_extensions import get_origin, get_args
except ImportError:                                         # pragma: no cover
    from typing import Annotated, get_origin, get_args
    from typing import _tp_cache                            # noqa

from .enums import EByteOrder


__all__ = ['T', 'TypeParams', 'is_annotated']


_DTYPE_RE = re.compile(
    r'^(?P<byteorder>[<|>])?'
    r'(?P<type>[?bBiufcmMUVOSat])'
    r'(?P<size>\d+)?$')


FieldTypes = Type[Union[bool, int, float, complex, bytes, str]]


class TypeParams(NamedTuple):
    byteorder: Optional[EByteOrder]
    type: FieldTypes
    size: Optional[int]
    signed: Optional[bool]

    def __repr__(self):
        byteorder = self.byteorder
        byteorder = repr(byteorder) if byteorder is not None else byteorder
        size = str(self.size) if self.size is not None else self.size
        return (
            f'{self.__class__.__name__}(byteorder={byteorder}, '
            f'type={self.type.__name__!r}, size={size}, signed={self.signed})'
        )


def str_to_type_params(typestr: str) -> TypeParams:
    """Convert a string describing a data type into type parameters.

    The ``typestr`` parameter is a string describing a data type.

    The *typestr* string format consists of 3 parts:

    * an (optional) character describing the byte order of the data

      - ``<``: little-endian,
      - ``>``: big-endian,
      - ``|``: not-relevant

    * a character code giving the basic type of the array, and
    * an integer providing the number of bytes the type uses

    The basic type character codes are:

    * ``i``: sighed integer
    * ``u``: unsigned integer
    * ``f``: float
    * ``c``: complex
    * ``S``: bytes (string)

    .. note:: *typestr* the format described above is a sub-set of the
       one used in the numpy "array interface".

    .. seealso:: https://numpy.org/doc/stable/reference/arrays.dtypes.html
       and https://numpy.org/doc/stable/reference/arrays.interface.html
    """
    mobj = _DTYPE_RE.match(typestr)
    if mobj is None:
        raise ValueError(f'invalid data type specifier: "{typestr}"')
    byteorder = mobj.group('byteorder')
    stype = mobj.group('type')
    size = mobj.group('size')
    signed = None

    if size is not None:
        size = int(size)
        if size <= 0:
            raise ValueError(f'invalid size: "{size}"')

    if byteorder == '|':
        byteorder = None
    elif byteorder is not None:
        byteorder = EByteOrder(byteorder)

    # if stype == '?' or (stype == 'b' and size == 1):
    #     type_ = bool
    # elif stype in 'bB':
    #     type_ = bytes
    # elif stype == 'i':
    if stype == 'i':
        type_ = int
        signed = True
    elif stype == 'u':
        type_ = int
        signed = False
    elif stype == 'f':
        type_ = float
    elif stype == 'c':
        type_ = complex
    # elif stype == 'm':
    #     type_ = datetime.timedelta
    # elif stype == 'M':
    #     type_ = datetime.datetime
    # elif stype == 'U':
    #     type_ = str
    elif stype == 'S':
        type_ = bytes
    # elif stype == 'V':
    #     type_ = bytes
    else:
        # '?': bool
        # 'b': (signed) byte (single item)
        # 'B': (unsigned) byte (single item)
        # 't': bitfield
        # 'O': object
        # 'U': (unicode) str (32bit UCS4 encoding)
        # 'a' : null terminated strings
        # 'm', 'M': timedelta and datetime
        raise TypeError(
            f'type specifier "{stype}" is valid for the "array protocol" but '
            f'not supported by bpack')

    return TypeParams(byteorder, type_, size, signed)


class T:
    """Allow to specify numeric type annotations using string descriptors.

    Example::

      >>> T['u4']                           # doctest: +NORMALIZE_WHITESPACE
      typing.Annotated[int, TypeParams(byteorder=None,
                       type='int', size=4, signed=False)]

    The resulting type annotation is a :class:`typing.Annotated` numeric type
    with attached a :class:`bpack.typing.TypeParams` instance.

    String descriptors, or *typestr*, are compatible with numpy (a sub-set
    of one used in the numpy "array interface").

    .. seealso:: :func:`str_to_type_params`, :class:`TypeParams`,
       https://numpy.org/doc/stable/reference/arrays.dtypes.html and
       https://numpy.org/doc/stable/reference/arrays.interface.html
    """

    __slots__ = ()

    def __new__(cls, *args, **kwargs):
        raise TypeError(f'Type "{cls.__name__}" cannot be instantiated.')

    @_tp_cache
    def __class_getitem__(cls, params):
        if not isinstance(params, str):
            raise TypeError(
                f'{cls.__name__}[...] should be used with a single argument '
                f'(a string describing a basic numeric type).')
        typestr = params
        metadata = str_to_type_params(typestr)
        return Annotated[metadata.type, metadata]

    def __init_subclass__(cls, *args, **kwargs):
        raise TypeError(f'Cannot subclass {cls.__module__}.{cls.__name__}')


def is_annotated(type_: Type) -> bool:
    """Return True if the input is an annotated numeric type.

    An *annotated numeric type* is assumed to be a :class:`typing.Annotated`
    type annotation of a basic numeric type with attached a
    :class:`bpack.typing.TypeParams` instance.

    .. seealso:: :class:`bpack.typing.T`.
    """
    if get_origin(type_) is Annotated:
        args = get_args(type_)
        if len(args) == 2:
            etype, params = args
            return isinstance(etype, type) and isinstance(params, TypeParams)
    return False



"""Base classes and utility functions for codecs."""

import abc
from typing import Optional, Type, Union

import bpack.utils
import bpack.descriptors

from .enums import EBaseUnits
from .descriptors import field_descriptors


CODEC_ATTR_NAME = '__bpack_decoder__'


class BaseCodec:
    baseunits: EBaseUnits

    def __init__(self, descriptor):
        self._descriptor = descriptor

        if bpack.baseunits(descriptor) is not self.baseunits:
            raise ValueError(
                f'{self.__class__.__module__}.{self.__class__.__name__} '
                f'only accepts descriptors with base units '
                f'"{self.baseunits.value}"')

    @property
    def descriptor(self):
        return self._descriptor


class Decoder(BaseCodec, abc.ABC):
    @abc.abstractmethod
    def decode(self, data: bytes):
        pass


class Encoder(BaseCodec, abc.ABC):
    @abc.abstractmethod
    def encode(self, record) -> bytes:
        pass


class Codec(Decoder, Encoder, abc.ABC):
    pass


CodecType = Union[Decoder, Encoder, Codec]


def make_codec_decorator(codec_type: Type[CodecType]):
    """Generate a codec decorator for the input decoder class."""
    @bpack.utils.classdecorator
    def codec(cls):
        """Class decorator to add (de)coding methods to a descriptor class.

        The decorator automatically generates a *Codec* object form the
        input descriptor class and attach to it methods for conversion
        form/to bytes.
        """
        codec_ = codec_type(descriptor=cls)
        bpack.utils.set_new_attribute(cls, CODEC_ATTR_NAME, codec_)

        if isinstance(codec_, Decoder):
            decode_func = bpack.utils.create_fn(
                name='frombytes',
                args=('cls', 'data'),
                body=[f'return cls.{CODEC_ATTR_NAME}.decode(data)'],
            )
            decode_func = classmethod(decode_func)
            bpack.utils.set_new_attribute(cls, 'frombytes', decode_func)

        if isinstance(codec_, Encoder):
            encode_func = bpack.utils.create_fn(
                name='tobytes',
                args=('self',),
                body=[f'return self.{CODEC_ATTR_NAME}.encode(self)'],
            )
            bpack.utils.set_new_attribute(cls, 'tobytes', encode_func)

        return cls

    return codec


def has_codec(descriptor,
              codec_type: Optional[Type[CodecType]] = None) -> bool:
    """Return True if the input descriptor has a codec attached.

    A descriptor decorated with a *codec* decorator has an attached codec
    instance and "frombytes"/"tobytes" methods (depending on the kind of
    codec).

    The *codec_type* parameter can be used to query for specific codec
    features:

    * codec_type=None: return True for any king of codec
    * codec_type=:class:`Decoder`: return True if the attached coded has
      decoding capabilities
    * codec_type=:class:`Encoder`: return True if the attached coded has
      encoding capabilities
    * codec_type=:class:`Codec`: return True if the attached coded has
      both encoding and decoding capabilities
    """
    if hasattr(descriptor, CODEC_ATTR_NAME):
        assert isinstance(get_codec(descriptor), (Codec, Decoder, Encoder))
        if codec_type is None:
            return True
        elif issubclass(codec_type, Codec):
            return (hasattr(descriptor, 'frombytes') and
                    hasattr(descriptor, 'tobytes'))
        elif issubclass(codec_type, Decoder):
            return hasattr(descriptor, 'frombytes')
        elif issubclass(codec_type, Encoder):
            return hasattr(descriptor, 'tobytes')
    return False


def get_codec(descriptor) -> CodecType:
    """Return the codec instance attached to the input descriptor."""
    return getattr(descriptor, CODEC_ATTR_NAME, None)


def get_codec_type(descriptor) -> Type[CodecType]:
    """Return the type of the codec attached to the input descriptor."""
    codec_ = getattr(descriptor, CODEC_ATTR_NAME, None)
    if codec_ is not None:
        return type(codec_)


def get_sequence_groups(descriptor):
    """Return slices to group values belonging to sequence fields.

    If the descriptor contains sequence fields this function returns a
    list of slices that can be used to extract values belonging to
    sequence fields from a flat list of all decoded values of the
    record.

    An empty list is returned if no sequence field is present in the
    descriptor.
    """
    groups = []
    offset = 0
    for descr in field_descriptors(descriptor):
        if bpack.is_descriptor(descr.type):
            nfields = len(bpack.fields(descr.type))                     # noqa
            slice_ = slice(offset, offset + nfields)

            def to_record(values, func=descr.type):
                return func(*values)

            groups.append((to_record, slice_))
            offset += nfields
        elif descr.repeat is not None:
            sequence_type = bpack.utils.sequence_type(descr.type, error=True)
            slice_ = slice(offset, offset + descr.repeat)
            groups.append((sequence_type, slice_))
            offset += descr.repeat
        else:
            offset += 1
    return groups



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import os
import sys
sys.path.insert(0, os.path.abspath('..'))

from setup import cfg                                           # noqa: E402


# -- Project information -----------------------------------------------------

project = 'bpack'
copyright = '2020-2021, Antonio Valentino'
author = 'Antonio Valentino'

# The full version, including alpha/beta/rc tags
release = cfg['version']

master_doc = 'index'


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.doctest',
    'sphinx.ext.intersphinx',
    'sphinx.ext.todo',
    'sphinx.ext.coverage',
    'sphinx.ext.ifconfig',
    'sphinx.ext.viewcode',
    'sphinx_rtd_theme',
]

try:
    import sphinxcontrib.spelling                               # noqa: F401
except ImportError:
    pass
else:
    extensions.append('sphinxcontrib.spelling')

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "sphinx_rtd_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# theme configuration
html_theme_options = {
    # 'vcs_pageview_mode': 'blob',
}

# theme context
html_context = {
    # 'github_url': 'https://github.com/avalentino/bpack/',
    'display_github': True,
    'github_user': 'avalentino',
    'github_repo': 'bpack',
    'github_version': 'main',
    'conf_py_path': '/docs/',  # Path in the checkout to the docs root
}

html_last_updated_fmt = ''


# -- Options for LaTeX output ------------------------------------------------
latex_documents = [
    # (startdocname, targetname, title, author, theme, toctree_only)
    (project, project + '.tex', 'Binary data structures (un-)Packing library',
     author, 'manual', False),
]

latex_domain_indices = False

latex_elements = {
    # 'papersize': 'a4paper',
    'pointsize': '12pt',
}

# -- Extension configuration -------------------------------------------------

# -- Options for autodoc extension -------------------------------------------
# autoclass_content = 'both'
autodoc_member_order = 'groupwise'
# autodoc_default_options = {}
autodoc_mock_imports = ['bitarray', 'bitstruct', 'numpy']

# -- Options for intersphinx extension ---------------------------------------

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    'https://docs.python.org/3/': None,
    'https://numpy.org/doc/stable/': None,
}

# -- Options for todo extension ----------------------------------------------

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True



"""Binary data structures (un-)Packing library.

bpack provides tools to describe, in a *declarative* way, and
encode/decode binary data.
"""

from .descriptors import (                                      # noqa: F401
    descriptor, field, fields, asdict, is_descriptor, is_field,
    baseunits, byteorder, bitorder, calcsize,
)
from .enums import EBaseUnits, EByteOrder, EBitOrder            # noqa: F401
from .typing import T                                           # noqa: F401


__version__ = '0.6.0'



"""Specific tests for the bitarray based decoder."""

from typing import List, Sequence
import pytest
import bpack


bpack_np = pytest.importorskip('bpack.np')


def test_sequence():
    backend = bpack_np
    bitorder = None
    size = 1
    repeat = 2
    encoded_data = bytes([3, 3, 4, 4])

    @backend.decoder
    @bpack.descriptor(baseunits=backend.Decoder.baseunits, bitorder=bitorder)
    class Record:
        field_1: List[int] = bpack.field(size=size, repeat=repeat)
        field_2: Sequence[int] = bpack.field(size=size, repeat=repeat)

    ref_record = Record([3, 3], (4, 4))
    record = Record.frombytes(encoded_data)
    assert list(record.field_1) == list(ref_record.field_1)
    assert list(record.field_2) == list(ref_record.field_2)

    for field, sequence_type in zip(bpack.fields(Record),
                                    (List[int], Sequence[int])):
        assert field.type == sequence_type



"""Utility functions and classes."""

import enum
import typing
import functools
import dataclasses
import collections.abc
from typing import Type, Union

from .typing import is_annotated
from .typing import get_origin, get_args, Annotated                     # noqa


def classdecorator(func):
    """Class decorator that can be used with or without parameters."""
    @functools.wraps(func)
    def wrapper(cls=None, **kwargs):
        def wrap(klass):
            return func(klass, **kwargs)

        # Check if called as @decorator or @decorator(...).
        if cls is None:
            # Called with parentheses
            return wrap

        # Called as @decorator without parentheses
        return wrap(cls)

    return wrapper


def create_fn(name, args, body, *, globals=None, locals=None,           # noqa
              return_type=dataclasses.MISSING):
    """Create a function object."""
    return dataclasses._create_fn(name, args, body,                     # noqa
                                  globals=globals, locals=locals,
                                  return_type=return_type)


def set_new_attribute(cls, name, value):
    """Programmatically add a new attribute/method to a class."""
    return dataclasses._set_new_attribute(cls, name, value)             # noqa


def sequence_type(type_: Type, error: bool = False) -> Union[Type, None]:
    """Return the sequence type associated to a typed sequence.

    The function return :class:`list` or :class:`tuple` if the input is
    considered a valid typed sequence, ``None`` otherwise.

    Please note that fields annotated with :class:`typing.Tuple` are not
    considered homogeneous sequences even if all items are specified to
    have the same type.
    """
    sequence_type_ = get_origin(type_)
    if sequence_type_ is None:
        return None
    if not issubclass(sequence_type_, typing.Sequence):
        return None
    args = get_args(type_)
    if len(args) < 1:
        return None
    if len(args) > 1:
        if error:
            raise TypeError(f'{type_} is not supported')
        else:
            return None

    if not is_annotated(args[0]) and not isinstance(args[0], type):
        # COMPATIBILITY: with typing_extensions and Python v3.7
        # need to be a concrete type
        return None                                         # pragma: no cover

    if not issubclass(sequence_type_, collections.abc.MutableSequence):
        sequence_type_ = tuple

    assert sequence_type_ in {list, tuple}

    return sequence_type_


def is_sequence_type(type_: Type, error: bool = False) -> bool:
    """Return True if the input is an homogeneous typed sequence.

    Please note that fields annotated with :class:`typing.Tuple` are not
    considered homogeneous sequences even if all items are specified to
    have the same type.
    """
    seq_type = sequence_type(type_, error=error)
    return seq_type is not None


def is_enum_type(type_: Type) -> bool:
    """Return True if the input is and :class:`enum.Enum`."""
    return get_origin(type_) is None and issubclass(type_, enum.Enum)


def enum_item_type(enum_cls: Type[enum.Enum]) -> Type:
    """Return the type of the items of an enum.Enum.

    This function also checks that all items of an enum have the same
    (or compatible) type.
    """
    if not is_enum_type(enum_cls):
        raise TypeError(f'"{enum_cls}" is not an enum. Enum')
    elif issubclass(enum_cls, int):
        return int
    else:
        types_ = [type(item.value) for item in enum_cls]
        type_ = types_.pop()
        for item in types_:
            if issubclass(item, type_):
                continue
            elif issubclass(type_, item):
                type_ = item
            else:
                raise TypeError(
                    'only Enum with homogeneous values are supported')
        return type_


def effective_type(type_: Union[Type, Type[enum.Enum], Type],
                   keep_annotations: bool = False) -> Type:
    """Return the effective type.

    In case of enums or sequences return the item type.
    """
    origin = get_origin(type_)
    if origin is None:
        if type_ is not None and issubclass(type_, enum.Enum):
            etype = enum_item_type(type_)
        else:
            etype = type_
    elif origin is Annotated:  # TODO: check issubclass(origin, Annotated):
        if keep_annotations:
            etype = type_
        else:
            etype, _ = get_args(type_)
    elif not issubclass(origin, typing.Sequence):
        etype = type_
    elif issubclass(origin, typing.Tuple):
        etype = type_
    else:
        # is a sequence
        args = get_args(type_)
        assert len(args) == 1
        etype = effective_type(args[0], keep_annotations=keep_annotations)

    return etype


def is_int_type(type_: Type) -> bool:
    """Return true if the effective type is an integer."""
    if is_sequence_type(type_):
        etype = effective_type(type_)
        return issubclass(etype, int)
    else:
        return issubclass(type_, int)



"""Test bpack.typing."""

import string

import pytest

import bpack.typing


TYPE_CODES = 'iufcS'
UNSUPPORTED_TYPE_CODES = '?bBOatmMUV'
INVALID_TYPE_CODES = (
    set(string.printable) - set(TYPE_CODES + UNSUPPORTED_TYPE_CODES)
)


class TestStrToTypeParams:
    @staticmethod
    @pytest.mark.parametrize('byteorder', ['>', '<', '|', ''])
    def test_byteorder(byteorder):
        s = f'{byteorder}i4'
        if byteorder in ('', '|'):
            byteorder = None
        else:
            byteorder = bpack.EByteOrder(byteorder if byteorder != '|' else '')
        params = bpack.typing.str_to_type_params(s)
        assert params.byteorder is byteorder

    @staticmethod
    @pytest.mark.parametrize('size', [1, 2, 4, 8, 3, 120])
    def test_size(size):
        s = f'i{size}'
        params = bpack.typing.str_to_type_params(s)
        assert params.size == size
        assert isinstance(params.size, int)

    @staticmethod
    @pytest.mark.parametrize('size', [0, -1])
    def test_invalid_size(size):
        s = f'i{size}'
        with pytest.raises(ValueError):
            bpack.typing.str_to_type_params(s)

    @staticmethod
    def test_no_size():
        s = 'i'
        params = bpack.typing.str_to_type_params(s)
        assert params.size is None

    @staticmethod
    @pytest.mark.parametrize('typecode', INVALID_TYPE_CODES)
    def test_invalid_typecode(typecode):
        s = f'{typecode}4'
        with pytest.raises(ValueError, match='invalid'):
            bpack.typing.str_to_type_params(s)

    @staticmethod
    @pytest.mark.parametrize('typecode', UNSUPPORTED_TYPE_CODES)
    def test_unsupported_typecode(typecode):
        s = f'{typecode}4'
        with pytest.raises(TypeError, match='not supported'):
            bpack.typing.str_to_type_params(s)

    @staticmethod
    @pytest.mark.parametrize('byteorder', ['>', '<', '|', ''])
    @pytest.mark.parametrize('typecode', ['S'])
    @pytest.mark.parametrize('size', ['2', '4', '8'])
    def test_bytes_typecode(byteorder, typecode, size):
        s = f'{byteorder}{typecode}{size}'
        params = bpack.typing.str_to_type_params(s)
        assert params.type is bytes
        assert params.signed is None

    @staticmethod
    @pytest.mark.parametrize('byteorder', ['>', '<', '|', ''])
    @pytest.mark.parametrize('typecode', ['f'])
    @pytest.mark.parametrize('size', ['2', '4', '8'])
    def test_float_typecode(byteorder, typecode, size):
        s = f'{byteorder}{typecode}{size}'
        params = bpack.typing.str_to_type_params(s)
        assert params.type is float
        assert params.signed is None

    @staticmethod
    @pytest.mark.parametrize('byteorder', ['>', '<', '|', ''])
    @pytest.mark.parametrize('typecode', ['c'])
    @pytest.mark.parametrize('size', ['4', '8', '16'])
    def test_complex_typecode(byteorder, typecode, size):
        s = f'{byteorder}{typecode}{size}'
        params = bpack.typing.str_to_type_params(s)
        assert params.type is complex
        assert params.signed is None

    @staticmethod
    @pytest.mark.parametrize('byteorder', ['>', '<', '|', ''])
    @pytest.mark.parametrize('typecode', ['i', 'u'])
    @pytest.mark.parametrize('size', ['2', '4', '8'])
    def test_int_typecode(byteorder, typecode, size):
        s = f'{byteorder}{typecode}{size}'
        signed = bool(typecode == 'i')
        params = bpack.typing.str_to_type_params(s)
        assert params.type is int
        assert params.signed == signed


@pytest.mark.parametrize(
    'typestr, type_, params',
    [('i4', int, bpack.typing.TypeParams(None, int, 4, True)),
     ('u2', int, bpack.typing.TypeParams(None, int, 2, False)),
     ('f8', float, bpack.typing.TypeParams(None, float, 8, None)),
     ('c16', complex, bpack.typing.TypeParams(None, complex, 16, None)),
     ('S128', bytes, bpack.typing.TypeParams(None, bytes, 128, None)),
     ('>i8', int, bpack.typing.TypeParams(bpack.EByteOrder.BE, int, 8, True)),
     ('f', float, bpack.typing.TypeParams(None, float, None, None))],
    ids=['i4', 'u2', 'f8', 'c16', 'S128', '>i8', 'f'])
def test_type_annotation(typestr, type_, params):
    T = bpack.typing.T[typestr]                                         # noqa
    assert isinstance(T(), type_)
    atype, metadata = bpack.typing.get_args(T)
    assert metadata == params
    assert atype == type_ == metadata.type



"""Bitarray based codec for binary data structures."""

import struct

from typing import Any, Callable, Optional

import bitarray
from bitarray.util import ba2int

import bpack
import bpack.utils
import bpack.codecs

from .enums import EBaseUnits, EBitOrder, EByteOrder
from .descriptors import field_descriptors


__all__ = ['Decoder', 'decoder', 'BACKEND_NAME', 'BACKEND_TYPE']


BACKEND_NAME = 'bitarray'
BACKEND_TYPE = EBaseUnits.BITS


FactoryType = Callable[[bitarray.bitarray], Any]


def ba_to_float_factory(size, byteorder: str = '>',
                        bitorder: str = 'big') -> FactoryType:
    """Convert a bitarray into a float."""
    assert bitorder == 'big'

    if size == 16:
        fmt = f'{byteorder}e'
    elif size == 32:
        fmt = f'{byteorder}f'
    elif size == 64:
        fmt = f'{byteorder}d'
    else:
        raise ValueError('floating point item size must be 16, 32 or 64 bits')

    codec = struct.Struct(fmt)

    def func(ba):
        return codec.unpack(ba.tobytes())[0]

    return func


def converter_factory(type_, size: Optional[int] = None, signed: bool = False,
                      byteorder: str = '>',
                      bitorder: str = 'big') -> FactoryType:
    if bpack.utils.is_sequence_type(type_, error=True):
        raise TypeError(
            f'backend "{BACKEND_NAME}" does not supports sequence types: '
            f'"{type_}".')
    etype = bpack.utils.effective_type(type_)
    if etype is int:
        def func(ba):
            return ba2int(ba, signed)
    elif etype is float:
        func = ba_to_float_factory(size, byteorder, bitorder)
    elif etype is bytes:
        def func(ba):
            return ba.tobytes()
    elif etype is str:
        def func(ba):
            return ba.tobytes().decode('ascii')
    elif etype is bool:
        def func(ba):
            return bool(bitarray.util.ba2int(ba))
    else:
        raise TypeError(
            f'type "{type_}" is not supported by the {__name__} backend'
            f'({BACKEND_NAME})')

    if etype is not type_:
        def converter(x, conv_func=func):
            return type_(conv_func(x))
    else:
        converter = func
    return converter


def _bitorder_to_baorder(bitorder: EBitOrder) -> str:
    if bitorder in {EBitOrder.MSB, EBitOrder.DEFAULT}:
        s = 'big'
    elif bitorder is EBitOrder.LSB:
        s = 'little'
    else:
        raise ValueError(f'invalid bit order: "{bitorder}"')
    return s


class Decoder(bpack.codecs.Decoder):
    """Bitarray based data decoder.

    Only supports "big endian" byte-order and MSB bit-order.
    """

    baseunits = EBaseUnits.BITS

    def __init__(self, descriptor, converters=converter_factory):
        """Initializer.

        The *descriptor* parameter* is a bpack record descriptor.
        """
        super().__init__(descriptor)

        byteorder = bpack.byteorder(descriptor)
        if byteorder in {EByteOrder.LE, EByteOrder.NATIVE}:
            raise NotImplementedError(
                f'byte order "{byteorder}" is not supported by the {__name__} '
                f'backend ({BACKEND_NAME})')

        bitorder = _bitorder_to_baorder(bpack.bitorder(descriptor))
        if bitorder != 'big':
            raise NotImplementedError(
                f'bit order "{bitorder}" is not supported by the {__name__} '
                f'backend ({BACKEND_NAME})')

        if callable(converters):
            conv_factory = converters
            byteorder_str = byteorder.value if byteorder.value else '>'
            converters = [
                conv_factory(field_descr.type, field_descr.size,
                             field_descr.signed, byteorder_str)
                for field_descr in field_descriptors(descriptor)
            ]

        if converters is not None:
            converters = list(converters)
            n_fields = len(list(bpack.fields(descriptor)))
            if len(converters) != n_fields:
                raise ValueError(
                    f'the number of converters ({len(converters)}) does not '
                    f'match the number of fields ({n_fields})')

        self._converters = converters
        self._slices = [
            slice(field_descr.offset, field_descr.offset + field_descr.size)
            for field_descr in field_descriptors(descriptor)
        ]

    def decode(self, data: bytes):
        """Decode binary data and return a record object."""
        ba = bitarray.bitarray()
        ba.frombytes(data)
        values = [ba[slice_] for slice_ in self._slices]

        if self._converters is not None:
            values = [
                convert(value) if convert is not None else value
                for convert, value in zip(self._converters, values)
            ]

        return self.descriptor(*values)


decoder = bpack.codecs.make_codec_decorator(Decoder)



"""Test bpack (record) descriptors."""

import dataclasses
from typing import List, Sequence, Tuple

import pytest

import bpack
import bpack.descriptors
from bpack import EBaseUnits, EByteOrder, EBitOrder
from bpack.descriptors import Field as BPackField


def test_base():
    @bpack.descriptor
    class Record:
        field_1: int = bpack.field(size=4, default=0)
        field_2: float = bpack.field(size=8, default=1/3)

    assert dataclasses.is_dataclass(Record)
    assert len(bpack.fields(Record)) == 2
    assert bpack.baseunits(Record) is EBaseUnits.BYTES  # default
    assert all(isinstance(f, BPackField) for f in bpack.fields(Record))


def test_frozen():
    @bpack.descriptor(frozen=True)
    class Record:
        field_1: int = bpack.field(size=4, default=0)
        field_2: float = bpack.field(size=8, default=1/3)

    assert dataclasses.is_dataclass(Record)
    assert len(bpack.fields(Record)) == 2
    assert bpack.baseunits(Record) is EBaseUnits.BYTES  # default
    assert all(isinstance(f, BPackField) for f in bpack.fields(Record))


def test_dataclass():
    with pytest.warns(DeprecationWarning):
        @bpack.descriptor
        @dataclasses.dataclass
        class Record:                                                   # noqa
            field_1: int = bpack.field(size=8, default=0)
            field_2: float = bpack.field(size=8, default=1/3)


@pytest.mark.parametrize(argnames='baseunits',
                         argvalues=[EBaseUnits.BYTES, EBaseUnits.BITS,
                                    'bits', 'bytes'])
def test_base_units(baseunits):
    @bpack.descriptor(baseunits=baseunits)
    class Record:
        field_1: int = bpack.field(size=8, default=0)
        field_2: float = bpack.field(size=8, default=1/3)

    assert bpack.baseunits(Record) is EBaseUnits(baseunits)
    assert bpack.baseunits(Record()) is EBaseUnits(baseunits)


def test_byte_alignment_warning():
    with pytest.warns(UserWarning,
                      match='bit struct not aligned to bytes'):
        @bpack.descriptor(baseunits=EBaseUnits.BITS)
        class Record:                                                   # noqa
            field_1: int = bpack.field(size=4, default=0)
            field_2: float = bpack.field(size=8, default=1/3)


def test_attrs():
    @bpack.descriptor
    class Record:
        field_1: int = bpack.field(size=4, default=0)
        field_2: float = bpack.field(size=8, default=1/3)

    assert hasattr(Record, bpack.descriptors.BASEUNITS_ATTR_NAME)
    assert hasattr(Record, bpack.descriptors.BYTEORDER_ATTR_NAME)
    assert hasattr(Record, bpack.descriptors.BITORDER_ATTR_NAME)
    assert hasattr(Record, bpack.descriptors.SIZE_ATTR_NAME)


@pytest.mark.parametrize(argnames='baseunits', argvalues=[None, 8, 'x'])
def test_invalid_baseunits(baseunits):
    with pytest.raises(ValueError):
        @bpack.descriptor(baseunits=baseunits)
        class Record:                                                   # noqa
            field_1: int = bpack.field(size=4, default=0)
            field_2: float = bpack.field(size=8, default=1/3)


@pytest.mark.parametrize(argnames='order',
                         argvalues=[EByteOrder.LE, EByteOrder.BE,
                                    EByteOrder.NATIVE, EByteOrder.DEFAULT,
                                    '<', '>', '=', ''])
def test_byteorder(order):
    @bpack.descriptor(byteorder=order)
    class Record:
        field_1: int = bpack.field(size=8, default=0)
        field_2: float = bpack.field(size=8, default=1/3)

    if isinstance(order, str):
        assert bpack.byteorder(Record) is EByteOrder(order)
        assert bpack.byteorder(Record()) is EByteOrder(order)
    else:
        assert bpack.byteorder(Record) is order
        assert bpack.byteorder(Record()) is order


@pytest.mark.parametrize(argnames='order', argvalues=['invalid', None])
def test_invalid_byteorder(order):
    with pytest.raises(ValueError):
        @bpack.descriptor(byteorder=order)
        class Record:                                                   # noqa
            field_1: int = bpack.field(size=8, default=0)


@pytest.mark.parametrize(argnames='order',
                         argvalues=[EBitOrder.LSB, EBitOrder.MSB,
                                    EBitOrder.DEFAULT, '<', '>', ''])
def test_bitorder(order):
    @bpack.descriptor(bitorder=order, baseunits=EBaseUnits.BITS)
    class Record:
        field_1: int = bpack.field(size=8, default=0)
        field_2: float = bpack.field(size=8, default=1/3)

    if isinstance(order, str):
        assert bpack.bitorder(Record) is EBitOrder(order)
        assert bpack.bitorder(Record()) is EBitOrder(order)
    else:
        assert bpack.bitorder(Record) is order
        assert bpack.bitorder(Record()) is order


def test_default_bitorder():
    @bpack.descriptor(baseunits=EBaseUnits.BITS)
    class Record:
        field_1: int = bpack.field(size=8, default=0)
        field_2: float = bpack.field(size=8, default=1 / 3)

    assert bpack.bitorder(Record) is EBitOrder.DEFAULT
    assert bpack.bitorder(Record()) is EBitOrder.DEFAULT


def test_invalid_bitorder():
    with pytest.raises(ValueError):
        @bpack.descriptor(byteorder='invalid', baseunits=EBaseUnits.BITS)
        class Record:                                                   # noqa
            field_1: int = bpack.field(size=8, default=0)


def test_bitorder_in_byte_descriptors():
    @bpack.descriptor(baseunits=EBaseUnits.BYTES)
    class Record:
        field_1: int = bpack.field(size=8, default=0)

    assert bpack.byteorder(Record) is EByteOrder.DEFAULT
    assert bpack.byteorder(Record()) is EByteOrder.DEFAULT

    with pytest.raises(ValueError):
        @bpack.descriptor(bitorder=EBitOrder.DEFAULT,
                          baseunits=EBaseUnits.BYTES)
        class Record:                                                   # noqa
            field_1: int = bpack.field(size=8, default=0)


def test_invalid_field_class():
    with pytest.raises(TypeError, match='size not specified'):
        @bpack.descriptor
        class Record:                                                   # noqa
            field_1: int = 0
            field_2: float = 1/3


def test_no_field_type():
    with pytest.raises(TypeError):
        @bpack.descriptor
        class Record:                                                   # noqa
            field_1 = bpack.field(size=4, default=0)


def test_invalid_field_size_type():
    with pytest.raises(TypeError):
        @bpack.descriptor
        class Record:                                                   # noqa
            field_1: int = bpack.field(size=4, default=0)
            field_2: float = bpack.field(size=None, default=1/3)


def test_inconsistent_field_offset_and_size():
    with pytest.raises(ValueError):
        @bpack.descriptor
        class Record:                                                   # noqa
            field_1: int = bpack.field(size=4, default=0)
            field_2: float = bpack.field(size=8, offset=1, default=1/3)


def test_inconsistent_field_type_and_signed():
    with pytest.warns(UserWarning, match='ignore'):
        @bpack.descriptor
        class Record:                                                   # noqa
            field_1: int = bpack.field(size=4, default=0)
            field_2: float = bpack.field(size=8, default=1/3, signed=True)


def test_repeat():
    @bpack.descriptor
    class Record:
        field_1: List[int] = bpack.field(size=4, default=0, repeat=2)
        field_2: float = bpack.field(size=8, default=1/3)

    assert bpack.calcsize(Record()) == 16


def test_no_repeat():
    with pytest.raises(TypeError):
        @bpack.descriptor
        class Record:                                                   # noqa
            field_1: List[int] = bpack.field(size=4, default=0)
            field_2: float = bpack.field(size=8, default=1 / 3)


def test_inconsistent_field_type_and_repeat():
    with pytest.raises(TypeError):
        @bpack.descriptor
        class Record:                                                   # noqa
            field_1: int = bpack.field(size=4, default=0, repeat=2)
            field_2: float = bpack.field(size=8, default=1/3, signed=True)


class TestExplicitSize:
    @staticmethod
    def test_explicit_size():
        size = 16

        @bpack.descriptor(size=size)
        class Record:
            field_1: int = bpack.field(size=4, default=0)
            field_2: float = bpack.field(size=8, default=1/3)

        assert bpack.calcsize(Record) == size
        assert bpack.calcsize(Record()) == size

    @staticmethod
    def test_invalid_explicit_size():
        with pytest.raises(ValueError):
            @bpack.descriptor(size=10)
            class Record:                                               # noqa
                field_1: int = bpack.field(size=4, default=0)
                field_2: float = bpack.field(size=8, default=1/3)

    @staticmethod
    def test_inconsistent_explicit_size_and_offset():
        with pytest.raises(bpack.descriptors.DescriptorConsistencyError):
            @bpack.descriptor(size=16)
            class Record:                                               # noqa
                field_1: int = bpack.field(size=4, default=0)
                field_2: float = bpack.field(size=8, default=1/3, offset=12)

    @staticmethod
    def test_inconsistent_explicit_size_and_repeat():
        with pytest.raises(bpack.descriptors.DescriptorConsistencyError):
            @bpack.descriptor(size=12)
            class Record:                                               # noqa
                field_1: List[int] = bpack.field(size=4, default=0, repeat=2)
                field_2: float = bpack.field(size=8, default=1/3)

    @staticmethod
    def test_inconsistent_explicit_size_offset_and_repeat():
        with pytest.raises(bpack.descriptors.DescriptorConsistencyError):
            @bpack.descriptor(size=20)
            class Record:                                               # noqa
                field_1: int = bpack.field(size=4, default=0)
                field_2: List[float] = bpack.field(size=8, default=1/3,
                                                   offset=8, repeat=2)

    @staticmethod
    def test_invalid_explicit_size_type():
        with pytest.raises(TypeError):
            @bpack.descriptor(size=10.5)
            class Record:                                               # noqa
                field_1: int = bpack.field(size=4, default=0)
                field_2: float = bpack.field(size=8, default=1/3)


class TestSize:
    # NOTE: basic tests are in test_descriptor_utils.test_calcsize.
    #       Here corner cases are addressed

    @staticmethod
    def test_len_with_offset_01():
        @bpack.descriptor
        class Record:
            field_1: int = bpack.field(size=4, default=0)
            field_2: float = bpack.field(size=8, offset=10, default=1/3)

        assert bpack.calcsize(Record) == 18

    @staticmethod
    def test_len_with_offset_02():
        @bpack.descriptor
        class Record:
            field_1: int = bpack.field(size=4, offset=10, default=0)
            field_2: float = bpack.field(size=8, default=1/3)

        assert bpack.calcsize(Record) == 22

    @staticmethod
    def test_len_with_offset_03():
        @bpack.descriptor
        class Record:
            field_1: int = bpack.field(size=4, offset=10, default=0)
            field_2: float = bpack.field(size=8, offset=20, default=1/3)

        assert bpack.calcsize(Record) == 28

    @staticmethod
    def test_len_with_offset_04():
        size = 30

        @bpack.descriptor(size=size)
        class Record:
            field_1: int = bpack.field(size=4, offset=10, default=0)
            field_2: float = bpack.field(size=8, offset=20, default=1/3)

        assert bpack.calcsize(Record) == size

    @staticmethod
    def test_len_with_repeat_01():
        @bpack.descriptor
        class Record:
            field_1: List[int] = bpack.field(size=4, default=0, repeat=2)
            field_2: float = bpack.field(size=8, default=1/3)

        assert bpack.calcsize(Record) == 16

    @staticmethod
    def test_len_with_repeat_02():
        @bpack.descriptor
        class Record:
            field_1: List[int] = bpack.field(size=4, default=0, repeat=2)
            field_2: List[float] = bpack.field(size=8, default=1/3, repeat=2)

        assert bpack.calcsize(Record) == 24

    @staticmethod
    def test_len_with_repeat_and_offset_01():
        @bpack.descriptor
        class Record:
            field_1: List[int] = bpack.field(size=4, default=0, repeat=2,
                                             offset=6)
            field_2: List[float] = bpack.field(size=8, default=1/3, repeat=2)

        assert bpack.calcsize(Record) == 30

    @staticmethod
    def test_len_with_repeat_and_offset_02():
        @bpack.descriptor
        class Record:
            field_1: List[int] = bpack.field(size=4, default=0, repeat=2)
            field_2: List[float] = bpack.field(size=8, default=1/3, repeat=2,
                                               offset=14)

        assert bpack.calcsize(Record) == 30

    @staticmethod
    def test_len_with_repeat_and_offset_03():
        with pytest.raises(bpack.descriptors.DescriptorConsistencyError):
            @bpack.descriptor
            class Record:                                               # noqa
                field_1: List[int] = bpack.field(size=4, default=0, repeat=2)
                field_2: List[float] = bpack.field(size=8, default=1/3,
                                                   repeat=2, offset=6)


def test_sequence_type():
    @bpack.descriptor
    class Record:
        field_1: List[int] = bpack.field(size=1, repeat=4)

    field_1 = bpack.fields(Record)[0]
    assert field_1.type is List[int]
    assert bpack.calcsize(Record) == 4

    @bpack.descriptor
    class Record:
        field_1: Sequence[int] = bpack.field(size=1, repeat=4)

    field_1 = bpack.fields(Record)[0]
    assert field_1.type is Sequence[int]

    with pytest.raises(TypeError):
        @bpack.descriptor
        class Record:                                                   # noqa
            field_1: Tuple[int, int] = bpack.field(size=1, repeat=2)

    with pytest.raises(TypeError):
        @bpack.descriptor
        class Record:                                                   # noqa
            field_1: Sequence = bpack.field(size=1, repeat=2)


def test_field_auto_size():
    @bpack.descriptor
    class Record:
        field_1: bool

    assert bpack.calcsize(Record) == 1

    with pytest.warns(UserWarning):
        @bpack.descriptor(baseunits=EBaseUnits.BITS)
        class Record:
            field_1: bool

    assert bpack.calcsize(Record) == 1


def test_nested_records():
    @bpack.descriptor
    class Record:
        field_1: int = bpack.field(size=4, default=11)
        field_2: float = bpack.field(size=4, default=22.22)

    @bpack.descriptor
    class NestedRecord:
        field_1: str = bpack.field(size=10, default='0123456789')
        field_2: Record = bpack.field(size=bpack.calcsize(Record),
                                      default=Record())
        field_3: int = bpack.field(size=4, default=3)

    record = Record()
    nested_record = NestedRecord()

    assert nested_record.field_1 == '0123456789'
    assert nested_record.field_2 == record
    assert nested_record.field_2.field_1 == record.field_1
    assert nested_record.field_2.field_2 == record.field_2
    assert nested_record.field_3 == 3

    with pytest.raises(bpack.descriptors.DescriptorConsistencyError):
        @bpack.descriptor
        class NestedRecord:                                             # noqa
            field_1: str = bpack.field(size=10, default='0123456789')
            field_2: Record = bpack.field(size=bpack.calcsize(Record) + 1,
                                          default=Record())
            field_3: int = bpack.field(size=4, default=3)


@pytest.mark.parametrize('baseunits',
                         [EBaseUnits.BYTES, EBaseUnits.BITS])
def test_nested_records_autosize(baseunits):
    @bpack.descriptor(baseunits=baseunits)
    class Record:
        field_1: int = bpack.field(size=4, default=1)
        field_2: int = bpack.field(size=4, default=2)

    @bpack.descriptor(baseunits=baseunits)
    class NestedRecord:
        field_1: int = bpack.field(size=4, default=1)
        field_2: Record = Record()  # auto-size (bytes)
        field_3: int = bpack.field(size=4, default=4)

    nested_record_size = bpack.calcsize(NestedRecord, baseunits)
    record_size = bpack.calcsize(Record, baseunits)
    assert nested_record_size == 8 + record_size



"""Test bpack field descriptors."""

import sys
import enum

from typing import List

import pytest

import bpack
from bpack.descriptors import get_field_descriptor


class TestFieldFactory:
    @staticmethod
    def test_base():
        bpack.field(size=1, offset=0, signed=False, default=0, repeat=1)

    @staticmethod
    def test_field_vs_field_class():
        field_ = bpack.field(size=1)
        assert bpack.descriptors.is_field(field_)
        assert isinstance(field_, bpack.descriptors.Field)

    @staticmethod
    @pytest.mark.parametrize(argnames='size', argvalues=[1.3, 'x'])
    def test_invalid_size_type(size):
        with pytest.raises(TypeError):
            bpack.field(size=size, default=1/3)

    @staticmethod
    @pytest.mark.parametrize(argnames='size', argvalues=[0, -8])
    def test_invalid_size(size):
        with pytest.raises(ValueError):
            bpack.field(size=size, default=1/3)

    @staticmethod
    @pytest.mark.parametrize(argnames='offset', argvalues=[1.3, 'x'])
    def test_invalid_offset_type(offset):
        with pytest.raises(TypeError):
            bpack.field(size=8, default=1/3, offset=offset)

    @staticmethod
    def test_invalid_offset():
        with pytest.raises(ValueError):
            bpack.field(size=8, default=1/3, offset=-8)

    @staticmethod
    @pytest.mark.parametrize('value', [-8, 'a'])
    def test_invalid_signed_type(value):
        with pytest.raises(TypeError):
            bpack.field(size=8, default=1, signed=value)                # noqa

    @staticmethod
    @pytest.mark.parametrize(argnames='repeat', argvalues=[1.3, 'x'])
    def test_invalid_repeat_type(repeat):
        with pytest.raises(TypeError):
            bpack.field(size=8, default=1/3, repeat=repeat)

    @staticmethod
    def test_invalid_repeat():
        with pytest.raises(ValueError):
            bpack.field(size=8, default=1/3, repeat=0)

    @staticmethod
    def test_metadata_key():
        field_ = bpack.field(size=1)
        assert bpack.descriptors.METADATA_KEY in field_.metadata


class TestRecordFields:
    @staticmethod
    def test_field_properties_01():
        @bpack.descriptor
        class Record:
            field_1: int = bpack.field(size=4, default=0, signed=True)
            field_2: float = bpack.field(size=8, default=1/3)
            field_3: List[int] = bpack.field(size=1, default=1, repeat=1)

        # name, type, size, offset, repeat
        field_data = [
            ('field_1', int, 4, 0, True, None),
            ('field_2', float, 8, 4, None, None),
            ('field_3', List[int], 1, 12, None, 1),
        ]

        for field_, data in zip(bpack.fields(Record), field_data):
            name, type_, size, offset, signed, repeat = data
            assert field_.name == name
            assert field_.type == type_
            field_descr = get_field_descriptor(field_)
            assert field_descr.type == type_
            assert field_descr.size == size
            assert field_descr.offset == offset
            assert field_descr.signed == signed
            assert field_descr.repeat == repeat

    @staticmethod
    def test_field_properties_02():
        @bpack.descriptor
        class Record:
            field_1: int = bpack.field(size=4, offset=1, default=0,
                                       signed=False)
            field_2: float = bpack.field(size=8, default=1/3)
            field_3: List[int] = bpack.field(size=1, default=1, repeat=1)

        # name, type, size, offset, repeat
        field_data = [
            ('field_1', int, 4, 1, False, None),
            ('field_2', float, 8, 5, None, None),
            ('field_3', List[int], 1, 13, None, 1),
        ]

        for field_, data in zip(bpack.fields(Record), field_data):
            name, type_, size, offset, signed, repeat = data
            assert field_.name == name
            assert field_.type == type_
            field_descr = get_field_descriptor(field_)
            assert field_descr.type == type_
            assert field_descr.size == size
            assert field_descr.offset == offset
            assert field_descr.signed == signed
            assert field_descr.repeat == repeat

    @staticmethod
    def test_field_properties_03():
        @bpack.descriptor
        class Record:
            field_1: int = bpack.field(size=4, offset=1, default=0)
            field_2: float = bpack.field(size=8, offset=6, default=1/3)

        # name, type, size, offset, repeat
        field_data = [
            ('field_1', int, 4, 1, None, None),
            ('field_2', float, 8, 6, None, None),
        ]

        for field_, data in zip(bpack.fields(Record), field_data):
            name, type_, size, offset, signed, repeat = data
            assert field_.name == name
            assert field_.type == type_
            field_descr = get_field_descriptor(field_)
            assert field_descr.type == type_
            assert field_descr.size == size
            assert field_descr.offset == offset
            assert field_descr.signed == signed
            assert field_descr.repeat == repeat

    @staticmethod
    def test_finvalid_field_type():
        with pytest.raises(TypeError):
            @bpack.descriptor
            class Record:                                       # noqa
                field_1: 'invalid' = bpack.field(size=4)        # noqa: F821


class TestEnumFields:
    @staticmethod
    def test_enum():
        class EEnumType(enum.Enum):
            A = 'a'
            B = 'b'
            C = 'c'

        @bpack.descriptor
        class Record:
            field_1: int = bpack.field(size=4, default=0)
            field_2: EEnumType = bpack.field(size=1, default=EEnumType.A)

        field_2 = bpack.fields(Record)[1]
        assert field_2.name == 'field_2'
        assert field_2.type is EEnumType
        assert field_2.default is EEnumType.A
        assert isinstance(Record().field_2, EEnumType)

    @staticmethod
    def test_int_enum():
        class EEnumType(enum.IntEnum):
            A = 1
            B = 2
            C = 4

        @bpack.descriptor
        class Record:
            field_1: int = bpack.field(size=4, default=0)
            field_2: EEnumType = bpack.field(size=1, signed=True,
                                             default=EEnumType.A)

        field_2 = bpack.fields(Record)[1]
        assert field_2.name == 'field_2'
        assert field_2.type is EEnumType
        assert field_2.default is EEnumType.A
        assert isinstance(Record().field_2, EEnumType)

    @staticmethod
    def test_intflag_enum():
        class EEnumType(enum.IntFlag):
            A = 1
            B = 2
            C = 4

        @bpack.descriptor
        class Record:
            field_1: int = bpack.field(size=4, default=0)
            field_2: EEnumType = bpack.field(size=1, default=EEnumType.A)

        field_2 = bpack.fields(Record)[1]
        assert field_2.name == 'field_2'
        assert field_2.type is EEnumType
        assert field_2.default is EEnumType.A
        assert isinstance(Record().field_2, EEnumType)

    @staticmethod
    def test_invalid_enum():
        class EEnumType(enum.Enum):
            A = 1
            B = 'b'
            C = 4

        with pytest.raises(TypeError):
            @bpack.descriptor
            class Record:                                               # noqa
                field_1: int = bpack.field(size=4, default=0)
                field_2: EEnumType = bpack.field(size=1, default=EEnumType.A)

    @staticmethod
    def test_invalid_signed_qualifier():
        class EEnumType(enum.Enum):
            A = 'a'
            B = 'b'
            C = 'c'

        with pytest.warns(UserWarning):
            @bpack.descriptor
            class Record:                                               # noqa
                field_1: int = bpack.field(size=4, default=0)
                field_2: EEnumType = bpack.field(size=1, signed=True,
                                                 default=EEnumType.A)


class TestFieldDescriptor:
    @staticmethod
    def test_empty_init():
        descr = bpack.descriptors.BinFieldDescriptor()
        assert descr.type is None
        assert descr.size is None
        assert descr.offset is None
        assert descr.signed is None
        assert descr.repeat is None
        assert len(vars(descr)) == 5

    @staticmethod
    def test_init():
        descr = bpack.descriptors.BinFieldDescriptor(int, 1, 2, True, 1)
        assert descr.type is int
        assert descr.size == 1
        assert descr.offset == 2
        assert descr.signed is True
        assert descr.repeat == 1
        assert len(vars(descr)) == 5

    @staticmethod
    def test_init_kw():
        descr = bpack.descriptors.BinFieldDescriptor(type=int, size=1,
                                                     offset=2, signed=True,
                                                     repeat=1)
        assert descr.type is int
        assert descr.size == 1
        assert descr.offset == 2
        assert descr.signed is True
        assert descr.repeat == 1
        assert len(vars(descr)) == 5

    @staticmethod
    def test_init_invalid_type():
        with pytest.raises(TypeError):
            bpack.descriptors.BinFieldDescriptor(size=1.1)              # noqa

        with pytest.raises(TypeError):
            bpack.descriptors.BinFieldDescriptor(offset=2.1)            # noqa

        with pytest.raises(TypeError):
            bpack.descriptors.BinFieldDescriptor(
                signed=complex(3.1, 0))                                 # noqa

        with pytest.raises(TypeError):
            bpack.descriptors.BinFieldDescriptor(repeat=1.1)            # noqa

    @staticmethod
    def test_init_invalid_value():
        with pytest.raises(ValueError):
            bpack.descriptors.BinFieldDescriptor(size=-1)

        with pytest.raises(ValueError):
            bpack.descriptors.BinFieldDescriptor(size=0)

        with pytest.raises(ValueError):
            bpack.descriptors.BinFieldDescriptor(offset=-1)

        with pytest.raises(ValueError):
            bpack.descriptors.BinFieldDescriptor(repeat=0)

    @staticmethod
    def test_validate():
        descr = bpack.descriptors.BinFieldDescriptor(int, 1)
        descr.validate()

        descr = bpack.descriptors.BinFieldDescriptor(int, 1, 2)
        descr.validate()

        descr = bpack.descriptors.BinFieldDescriptor(int, 1, 2, True)
        descr.validate()

        descr = bpack.descriptors.BinFieldDescriptor(List[int], 1, 2, True, 1)
        descr.validate()

    @staticmethod
    def test_validation_warning():
        descr = bpack.descriptors.BinFieldDescriptor(type=float, size=4,
                                                     signed=True)
        with pytest.warns(UserWarning, match='ignore'):
            descr.validate()

    @staticmethod
    def test_validation_error():
        descr = bpack.descriptors.BinFieldDescriptor()
        with pytest.raises(TypeError):
            descr.validate()

        descr = bpack.descriptors.BinFieldDescriptor(type=int)
        with pytest.raises(TypeError):
            descr.validate()

        descr = bpack.descriptors.BinFieldDescriptor(size=1)
        with pytest.raises(TypeError):
            descr.validate()

        descr = bpack.descriptors.BinFieldDescriptor(type=int, size=1,
                                                     repeat=2)
        with pytest.raises(TypeError):
            descr.validate()

    @staticmethod
    def test_post_validation_error_on_type():
        descr = bpack.descriptors.BinFieldDescriptor(int, 1, 2)
        descr.validate()
        descr.type = None
        with pytest.raises(TypeError):
            descr.validate()

    @staticmethod
    @pytest.mark.parametrize('size, error_type',
                             [(None, TypeError),
                              (0, ValueError),
                              (-1, ValueError),
                              (1.1, TypeError)],
                             ids=['None', 'zero', 'negative', 'float'])
    def test_post_validation_error_on_size(size, error_type):
        descr = bpack.descriptors.BinFieldDescriptor(int, 1, 2)
        descr.validate()
        descr.size = size
        with pytest.raises(error_type):
            descr.validate()

    @staticmethod
    @pytest.mark.parametrize('offset, error_type',
                             [(-1, ValueError),
                              (1.1, TypeError)],
                             ids=['negative', 'float'])
    def test_post_validation_error_on_offset(offset, error_type):
        descr = bpack.descriptors.BinFieldDescriptor(int, 1, 2)
        descr.validate()
        descr.offset = offset
        with pytest.raises(error_type):
            descr.validate()

    @staticmethod
    def test_post_validation_warning_on_signed():
        descr = bpack.descriptors.BinFieldDescriptor(int, 1, 2, signed=True)
        descr.validate()
        descr.type = float
        with pytest.warns(UserWarning, match='ignore'):
            descr.validate()

    @staticmethod
    def test_post_validation_error_on_repeat():
        descr = bpack.descriptors.BinFieldDescriptor(int, 1, 2, signed=True)
        descr.validate()
        descr.repeat = 2
        with pytest.raises(TypeError):
            descr.validate()

        descr = bpack.descriptors.BinFieldDescriptor(List[int], 1, 2,
                                                     signed=True, repeat=2)
        descr.validate()
        descr.repeat = 0
        with pytest.raises(ValueError):
            descr.validate()

    def test_methods(self):
        descr = bpack.descriptors.BinFieldDescriptor(int, 1)
        descr.validate()
        assert descr.is_int_type()
        assert not descr.is_sequence_type()
        assert not descr.is_enum_type()

        descr = bpack.descriptors.BinFieldDescriptor(float, 1)
        descr.validate()
        assert not descr.is_int_type()
        assert not descr.is_sequence_type()
        assert not descr.is_enum_type()

        descr = bpack.descriptors.BinFieldDescriptor(List[int], 1, repeat=10)
        descr.validate()
        assert descr.is_int_type()
        assert descr.is_sequence_type()
        assert not descr.is_enum_type()

        descr = bpack.descriptors.BinFieldDescriptor(List[float], 1, repeat=10)
        descr.validate()
        assert not descr.is_int_type()
        assert descr.is_sequence_type()
        assert not descr.is_enum_type()

        class EEnumType(enum.Enum):
            A = 'a'

        descr = bpack.descriptors.BinFieldDescriptor(EEnumType, 1)
        descr.validate()
        assert not descr.is_int_type()
        assert not descr.is_sequence_type()
        assert descr.is_enum_type()

        class EEnumType(enum.IntEnum):
            A = 1

        descr = bpack.descriptors.BinFieldDescriptor(EEnumType, 1)
        descr.validate()
        assert descr.is_int_type()
        assert not descr.is_sequence_type()
        assert descr.is_enum_type()

        class EEnumType(enum.IntFlag):
            A = 1

        descr = bpack.descriptors.BinFieldDescriptor(EEnumType, 1)
        descr.validate()
        assert descr.is_int_type()
        assert not descr.is_sequence_type()
        assert descr.is_enum_type()


class TestAnnotatedType:
    @staticmethod
    @pytest.mark.parametrize('byteorder', ['>', '<', '|', ''],
                             ids=['>', '<', '|', 'None'])
    def test_annotated_type(byteorder):
        @bpack.descriptor(byteorder=byteorder if byteorder != '|' else '')
        class Record:
            field_1: bpack.T[f'{byteorder}i4']                  # noqa: F821
            field_2: bpack.T[f'{byteorder}u4']                  # noqa: F821
            field_3: bpack.T[f'{byteorder}f4']                  # noqa: F821
            field_4: bpack.T[f'{byteorder}c4']                  # noqa: F821
            field_5: bpack.T[f'{byteorder}S4']                  # noqa: F821

        fields = dict(
            (field.name, get_field_descriptor(field))
            for field in bpack.fields(Record)
        )

        assert fields['field_1'].type == int
        assert fields['field_1'].size == 4
        assert fields['field_1'].signed is True
        assert fields['field_1'].repeat is None

        assert fields['field_2'].type == int
        assert fields['field_2'].size == 4
        assert fields['field_2'].signed is False
        assert fields['field_2'].repeat is None

        assert fields['field_3'].type == float
        assert fields['field_3'].size == 4
        assert fields['field_3'].signed is None
        assert fields['field_3'].repeat is None

        assert fields['field_4'].type == complex
        assert fields['field_4'].size == 4
        assert fields['field_4'].signed is None
        assert fields['field_4'].repeat is None

        assert fields['field_5'].type == bytes
        assert fields['field_5'].size == 4
        assert fields['field_5'].signed is None
        assert fields['field_5'].repeat is None

    @staticmethod
    def test_list_with_annotated_type():
        typestr = 'i4'

        @bpack.descriptor
        class Record:
            field_1: List[bpack.T[typestr]] = bpack.field(repeat=2)

        field = bpack.fields(Record)[0]
        assert field.type == List[bpack.T[typestr]]

        field_descr = get_field_descriptor(field)
        assert field_descr.type == List[int]
        assert field_descr.size == 4
        assert field_descr.signed is True
        assert field_descr.repeat == 2

    @staticmethod
    def test_byteorder_consistency():
        typestr = '>i8'
        with pytest.raises(bpack.descriptors.DescriptorConsistencyError):
            @bpack.descriptor(byteorder=bpack.EByteOrder.LE)
            class Record:                                       # noqa
                field: bpack.T[typestr]

        typestr = '<i8'
        with pytest.raises(bpack.descriptors.DescriptorConsistencyError):
            @bpack.descriptor(byteorder=bpack.EByteOrder.BE)   # noqa: F811
            class Record:                                       # noqa
                field: bpack.T[typestr]

        typestr = '>i8' if sys.byteorder == 'little' else '<i8'
        with pytest.raises(bpack.descriptors.DescriptorConsistencyError):
            @bpack.descriptor                                   # noqa: F811
            class Record:                                       # noqa
                field: bpack.T[typestr]

        typestr = '<i8' if sys.byteorder == 'little' else '>i8'

        @bpack.descriptor                                       # noqa: F811
        class Record:                                           # noqa
            field: bpack.T[typestr]

        typestr = '|i8'

        @bpack.descriptor(byteorder=bpack.EByteOrder.BE)       # noqa: F811
        class Record:                                           # noqa
            field: bpack.T[typestr]

    @staticmethod
    def test_size_consistency():
        typestr = 'i8'

        @bpack.descriptor
        class Record:
            field: bpack.T[typestr] = bpack.field(size=8, signed=True)

        descr = get_field_descriptor(bpack.fields(Record)[0])
        assert descr.type is int
        assert descr.size == 8
        assert descr.repeat is None
        assert descr.signed is True

        typestr = 'u8'

        @bpack.descriptor
        class Record:
            field: bpack.T[typestr] = bpack.field(size=8, signed=False)

        descr = get_field_descriptor(bpack.fields(Record)[0])
        assert descr.type is int
        assert descr.size == 8
        assert descr.repeat is None
        assert descr.signed is False

        typestr = 'f8'

        @bpack.descriptor
        class Record:
            field: bpack.T[typestr] = bpack.field(size=8)

        descr = get_field_descriptor(bpack.fields(Record)[0])
        assert descr.type is float
        assert descr.size == 8
        assert descr.repeat is None
        assert descr.signed is None

    @staticmethod
    def test_size_consistency_error():
        typestr = 'i8'
        with pytest.raises(bpack.descriptors.DescriptorConsistencyError):
            @bpack.descriptor
            class Record:                                               # noqa
                field: bpack.T[typestr] = bpack.field(size=3)

    @staticmethod
    def test_signed_consistency_error():
        typestr = 'i8'
        with pytest.raises(bpack.descriptors.DescriptorConsistencyError):
            @bpack.descriptor
            class Record:                                               # noqa
                field: bpack.T[typestr] = bpack.field(size=8, signed=False)

    @staticmethod
    @pytest.mark.parametrize('typestr',
                             ['invalid', '@i8', '|x8', '|i0', '|ii'])
    def test_invalid_typestr(typestr):
        with pytest.raises(ValueError):
            @bpack.descriptor
            class Record:                                               # noqa
                field: bpack.T[typestr] = bpack.field(size=1)



"""Specific tests for the bitarray based decoder."""

from typing import List, Sequence

import pytest

import bpack
bpack_ba = pytest.importorskip('bpack.ba')


@pytest.mark.parametrize(
    'size, data',
    [(16, bytes([0b00111100, 0b00000000])),
     (32, bytes([0b00111111, 0b10000000, 0b00000000, 0b00000000])),
     (64, bytes([0b00111111, 0b11110000, 0b00000000, 0b00000000,
                 0b00000000, 0b00000000, 0b00000000, 0b00000000]))],
    ids=['float16', 'float32', 'float64'])
def test_float(size, data):
    backend = bpack_ba

    @backend.decoder
    @bpack.descriptor(baseunits=bpack.EBaseUnits.BITS)
    class Record:
        field_1: float = bpack.field(size=size)

    record = Record.frombytes(data)
    assert record.field_1 == 1.


def test_invalid_float_size():
    backend = bpack_ba

    with pytest.raises(ValueError):
        @backend.decoder
        @bpack.descriptor(baseunits=bpack.EBaseUnits.BITS)
        class Record:                                                   # noqa
            field_1: float = bpack.field(size=80)


def test_little_endian():
    backend = bpack_ba

    with pytest.raises(NotImplementedError):
        @backend.decoder
        @bpack.descriptor(baseunits=bpack.EBaseUnits.BITS,
                          byteorder=bpack.EByteOrder.LE,
                          frozen=True)
        class Record:                                                   # noqa
            field_1: int = bpack.field(size=8, default=1)


def test_invalid_bitorder():
    backend = bpack_ba

    with pytest.raises(NotImplementedError):
        @backend.decoder
        @bpack.descriptor(baseunits=bpack.EBaseUnits.BITS,
                          bitorder=bpack.EBitOrder.LSB,
                          frozen=True)
        class Record:                                                   # noqa
            field_1: int = bpack.field(size=8, default=1)


def test_sequence():
    backend = bpack_ba

    with pytest.raises(TypeError):
        @backend.decoder
        @bpack.descriptor(baseunits=bpack.EBaseUnits.BITS)
        class Record:                                                   # noqa
            field_1: List[int] = bpack.field(size=4, signed=False,
                                             repeat=2, default=3)
            field_2: Sequence[int] = bpack.field(size=4, signed=False,
                                                 repeat=2, default=4)



"""Test bpack.utils."""

import enum
import typing

import pytest

import bpack
import bpack.utils


class TestEnumType:
    @staticmethod
    @pytest.mark.parametrize('value', [1, 1.1, None, 'a', b'a'])
    def test_enum_type(value):
        class EEnumType(enum.Enum):
            A = value

        assert bpack.utils.enum_item_type(EEnumType) is type(value)

    @staticmethod
    def test_invalid_enum_type():
        class EInvalidEnumType:
            A = 'a'

        with pytest.raises(TypeError):
            assert bpack.utils.enum_item_type(EInvalidEnumType)         # noqa

    @staticmethod
    def test_unsupported_enum_type():
        class EEnumType(enum.Enum):
            A = 'a'
            B = 'b'
            c = 1

        with pytest.raises(TypeError):
            assert bpack.utils.enum_item_type(EEnumType)

    @staticmethod
    def test_int_enum_type():
        class EEnumType(enum.IntEnum):
            A = 1

        assert bpack.utils.enum_item_type(EEnumType) is int

    @staticmethod
    def test_flag_enum_type():
        class EEnumType(enum.IntFlag):
            A = 1

        assert bpack.utils.enum_item_type(EEnumType) is int

    @staticmethod
    def test_subclass_types():
        class EFlagEnumType(enum.IntFlag):
            A = 1
            B = 2

        class EEnumType(enum.Enum):
            A = EFlagEnumType.A
            B = 0
            c = EFlagEnumType.B

        assert bpack.utils.enum_item_type(EEnumType) is int


def test_effective_type():
    assert bpack.utils.effective_type(bool) is bool
    assert bpack.utils.effective_type(int) is int
    assert bpack.utils.effective_type(str) is str
    assert bpack.utils.effective_type(bytes) is bytes
    assert bpack.utils.effective_type(float) is float
    assert bpack.utils.effective_type(None) is None                     # noqa

    for type_ in (typing.Type[typing.Any], typing.Tuple[int, float],
                  typing.Tuple[int]):
        assert bpack.utils.effective_type(type_) == type_


def test_effective_enum():
    class EStrEnumType(enum.Enum):
        A = 'a'
        B = 'b'

    assert bpack.utils.effective_type(EStrEnumType) is str

    class EBytesEnumType(enum.Enum):
        A = b'a'
        B = b'b'

    assert bpack.utils.effective_type(EBytesEnumType) is bytes

    class EIntEnumType(enum.Enum):
        A = 1
        B = 2

    assert bpack.utils.effective_type(EIntEnumType) is int

    class EFlagEnumType(enum.Enum):
        A = 1
        B = 2

    assert bpack.utils.effective_type(EFlagEnumType) is int


def test_effective_type_from_seq():
    assert bpack.utils.effective_type(typing.List[bool]) is bool
    assert bpack.utils.effective_type(typing.List[int]) is int
    assert bpack.utils.effective_type(typing.List[str]) is str
    assert bpack.utils.effective_type(typing.List[bytes]) is bytes
    assert bpack.utils.effective_type(typing.List[float]) is float


def test_effective_type_from_annotated_type():
    assert bpack.utils.effective_type(bpack.T['i']) is int
    assert bpack.utils.effective_type(bpack.T['u']) is int
    assert bpack.utils.effective_type(bpack.T['f']) is float
    assert bpack.utils.effective_type(bpack.T['c']) is complex
    assert bpack.utils.effective_type(bpack.T['S']) is bytes


def test_invalid_effective_type():
    with pytest.raises(TypeError):
        bpack.utils.effective_type('i8')                                # noqa


def test_effective_type_keep_annotations():
    atype = bpack.T['i8']
    etype = bpack.utils.effective_type(atype, keep_annotations=True)
    assert etype == atype


def test_get_sequence_type():
    type_ = typing.List[int]
    assert bpack.utils.sequence_type(type_) is list

    type_ = typing.Sequence[int]
    assert bpack.utils.sequence_type(type_) is tuple

    type_ = typing.Tuple[int, int]
    assert bpack.utils.sequence_type(type_) is None
    with pytest.raises(TypeError):
        bpack.utils.sequence_type(type_, error=True)

    assert bpack.utils.sequence_type(list) is None
    assert bpack.utils.sequence_type(typing.List) is None
    assert bpack.utils.sequence_type(typing.Sequence) is None
    assert bpack.utils.sequence_type(typing.Tuple) is None
    assert bpack.utils.sequence_type(typing.Type[typing.Any]) is None


def test_get_sequence_type_from_annotated_type():
    typestr = 'i8'

    type_ = typing.List[bpack.T[typestr]]
    assert bpack.utils.sequence_type(type_) is list

    type_ = typing.Sequence[bpack.T[typestr]]
    assert bpack.utils.sequence_type(type_) is tuple


def test_is_sequence_type():
    type_ = typing.List[int]
    assert bpack.utils.is_sequence_type(type_)

    type_ = typing.Sequence[int]
    assert bpack.utils.is_sequence_type(type_)

    type_ = typing.Tuple[int, int]
    assert not bpack.utils.is_sequence_type(type_)
    with pytest.raises(TypeError):
        bpack.utils.is_sequence_type(type_, error=True)

    assert not bpack.utils.is_sequence_type(list)
    assert not bpack.utils.is_sequence_type(typing.List)
    assert not bpack.utils.is_sequence_type(typing.Sequence)
    assert not bpack.utils.is_sequence_type(typing.Tuple)


def test_is_sequence_type_from_annotated_type():
    typestr = 'i8'

    type_ = typing.List[bpack.T[typestr]]
    assert bpack.utils.is_sequence_type(type_)

    type_ = typing.Sequence[bpack.T[typestr]]
    assert bpack.utils.is_sequence_type(type_)



"""Enumeration types for the bpack package."""

import sys
import enum


class EBaseUnits(enum.Enum):
    """Base units used to specify size and offset parameters in descriptors."""

    BITS = 'bits'
    BYTES = 'bytes'


class EByteOrder(enum.Enum):
    """Enumeration for byte order (endianess).

    .. note::

        the :data:`EByteOrder.DEFAULT` is equivalent to
        :data:`EByteOrder.NATIVE` for binary structures having
        :data:`EBaseUnits.BYTE` base units, and :data:`EByteOrder.BE`
        for binary structures having :data:`EBaseUnits.BIT` base units.
    """

    BE = '>'
    LE = '<'
    NATIVE = '='
    DEFAULT = ''

    @classmethod
    def get_native(cls):
        return cls.LE if sys.byteorder == 'little' else cls.BE


class EBitOrder(enum.Enum):
    """Enumeration for bit order."""

    MSB = '>'
    LSB = '<'
    DEFAULT = ''



"""Numpy based codec for binary data structures."""

import collections

import numpy as np

import bpack
import bpack.utils
import bpack.codecs

from .enums import EBaseUnits
from .descriptors import (
    field_descriptors, get_field_descriptor, BinFieldDescriptor,
)


__all__ = [
    'Decoder', 'decoder', 'BACKEND_NAME', 'BACKEND_TYPE',
    'descriptor_to_dtype',
]


BACKEND_NAME = 'numpy'
BACKEND_TYPE = EBaseUnits.BYTES


def bin_field_descripor_to_dtype(field_descr: BinFieldDescriptor) -> np.dtype:
    """Convert a field descriptor into a :class:`numpy.dtype`.

    .. seealso:: :class:`bpack.descriptors.BinFieldDescriptor`.
    """
    # TODO: add byteorder
    size = field_descr.size
    etype = bpack.utils.effective_type(field_descr.type)
    typecode = np.dtype(etype).kind

    if etype in (bytes, str):
        typecode = 'S'
    elif etype is int and not field_descr.signed:
        typecode = 'u'

    if typecode == 'O':
        raise TypeError(f'unsupported type: {field_descr.type:!r}')

    repeat = field_descr.repeat
    repeat = repeat if repeat and repeat > 1 else ''

    return np.dtype(f'{repeat}{typecode}{size}')


def descriptor_to_dtype(descriptor) -> np.dtype:
    """Convert the descriptor of a binary record into a :class:`numpy.dtype`.

    Please note that (unicode) strings are treated as "utf-8" encoded
    byte strings.
    UCS4 encoded strings are not supported.

    Sequences (:class:`typing.Sequence` and :class:`typing.List`) are
    always converted into :class:`numpy.ndarray`.

    .. seealso:: :func:`bpack.descriptors.descriptor`.
    """
    params = collections.defaultdict(list)
    for field in bpack.fields(descriptor):
        field_descr = get_field_descriptor(field)
        if bpack.is_descriptor(field_descr.type):
            dtype = descriptor_to_dtype(field_descr.type)
        else:
            dtype = bin_field_descripor_to_dtype(field_descr)
        params['names'].append(field.name)
        params['formats'].append(dtype)
        params['offsets'].append(field_descr.offset)
        # params['titles'].append('...')

    params = dict(params)  # numpy do not accept defaultdict
    params['itemsize'] = bpack.calcsize(descriptor)

    dt = np.dtype(dict(params))

    byteorder = bpack.byteorder(descriptor).value
    if byteorder:
        dt = dt.newbyteorder(byteorder)

    return dt


def _converter_factory(type_):
    etype = bpack.utils.effective_type(type_)
    if bpack.utils.is_enum_type(type_):
        if etype is str:
            def converter(x, cls=type_):
                return cls(x.tobytes().decode('utf-8'))
        else:
            def converter(x, cls=type_):
                return cls(x)
    elif etype is str:
        def converter(x):
            return x.tobytes().decode('utf-8')
    elif bpack.is_descriptor(type_):
        def converter(x, cls=type_):
            return cls(*x)
    else:
        converter = None

    return converter


class Decoder(bpack.codecs.Decoder):
    """Numpy based data decoder.

    (Unicode) strings are treated as "utf-8" encoded byte strings.
    UCS4 encoded strings are not supported.
    """

    baseunits = EBaseUnits.BYTES

    def __init__(self, descriptor):
        """Initializer.

        The *descriptor* parameter* is a bpack record descriptor.
        """
        super().__init__(descriptor)

        assert bpack.bitorder(descriptor) is None

        converters = [
            (idx, _converter_factory(field_descr.type))
            for idx, field_descr in enumerate(field_descriptors(descriptor))
        ]

        self._dtype = descriptor_to_dtype(descriptor)
        self._converters = [(idx, func) for idx, func in converters if func]

    @property
    def dtype(self):
        return self._dtype

    def decode(self, data: bytes, count: int = 1):
        """Decode binary data and return a record object."""
        v = np.frombuffer(data, dtype=self._dtype, count=count)
        if self._converters:
            out = []
            for item in v:
                item = list(item)
                for idx, func in self._converters:
                    item[idx] = func(item[idx])
                out.append(self.descriptor(*item))
        else:
            out = [self.descriptor(*item) for item in v]
        if len(v) == 1:
            out = out[0]
        return out


decoder = bpack.codecs.make_codec_decorator(Decoder)



"""Bitstruct based codec for binary data structures."""

from typing import Optional

import bitstruct

import bpack
import bpack.utils
import bpack.codecs

from .enums import EBaseUnits, EByteOrder
from .codecs import get_sequence_groups, has_codec, get_codec
from .descriptors import field_descriptors


__all__ = ['Decoder', 'decoder', 'BACKEND_NAME', 'BACKEND_TYPE']


BACKEND_NAME = 'bitstruct'
BACKEND_TYPE = EBaseUnits.BITS


class BitStruct(bitstruct.CompiledFormat):
    def __init__(self, format: str):                                    # noqa
        super().__init__(format)
        self._format: str = format

    @property
    def format(self) -> str:
        return self._format


_TYPE_TO_STR = {
    bool: 'b',
    int: 'u',
    (int, False): 'u',
    (int, True): 's',
    float: 'f',
    bytes: 'r',
    str: 't',
    None: 'p',
}


def _to_fmt(type_, size: int, bitorder: str = '',
            signed: Optional[bool] = None,
            repeat: Optional[int] = None) -> str:
    assert size > 0, f'invalid size: {size:r}'
    assert bitorder in ('', '>', '<'), f'invalid order: {bitorder:r}'
    if repeat is None:
        repeat = 1
    assert repeat > 0, f'invalid repeat: {repeat:r}'

    if has_codec(type_, bpack.codecs.Decoder):
        decoder_ = get_codec(type_)
        if isinstance(decoder_, Decoder):
            return decoder_.format
    elif (bpack.is_descriptor(type_) and
          bpack.baseunits(type_) is Decoder.baseunits):
        decoder_ = Decoder(type_)
        return decoder_.format

    etype = bpack.utils.effective_type(type_)
    key = (etype, signed) if etype is int and signed is not None else etype

    try:
        fmt = f'{bitorder}{_TYPE_TO_STR[key]}{size}' * repeat           # noqa
    except KeyError:
        raise TypeError(f'unsupported type: {etype:!r}')

    return fmt


def _endianess_to_str(order: EByteOrder) -> str:
    if order is EByteOrder.NATIVE:
        return EByteOrder.get_native().value
    return order.value


class Decoder(bpack.codecs.Decoder):
    """Bitstruct based data decoder.

    Default bit-order: MSB.
    """

    baseunits = EBaseUnits.BITS

    def __init__(self, descriptor):
        """Initializer.

        The *descriptor* parameter* is a bpack record descriptor.
        """
        super().__init__(descriptor)

        byteorder = bpack.byteorder(descriptor)
        byteorder = _endianess_to_str(byteorder)
        bitorder = bpack.bitorder(descriptor).value

        fmt = ''.join(
            _to_fmt(field_descr.type, size=field_descr.size, bitorder=bitorder,
                    signed=field_descr.signed, repeat=field_descr.repeat)
            for field_descr in field_descriptors(descriptor, pad=True)
        )
        fmt = fmt + byteorder  # byte order

        self._codec = BitStruct(fmt)
        self._converters = [
            (idx, field_descr.type)
            for idx, field_descr in enumerate(field_descriptors(descriptor))
            if bpack.utils.is_enum_type(field_descr.type)
        ]
        self._groups = get_sequence_groups(descriptor)

    @property
    def format(self) -> str:
        """Return the *bitstruct* format string."""
        return self._codec.format

    def decode(self, data: bytes):
        """Decode binary data and return a record object."""
        values = list(self._codec.unpack(data))

        for idx, func in self._converters:
            values[idx] = func(values[idx])

        for type_, slice_ in self._groups[::-1]:
            subtype = type_(values[slice_])
            del values[slice_]
            values.insert(slice_.start, subtype)

        return self.descriptor(*values)


decoder = bpack.codecs.make_codec_decorator(Decoder)



"""Descriptors for binary records."""

import copy
import enum
import math
import types
import warnings
import dataclasses
from typing import Iterable, Optional, Sequence, Type, Union

import bpack.typing
import bpack.utils
from .utils import classdecorator
from .enums import EBaseUnits, EByteOrder, EBitOrder

__all__ = [
    'descriptor', 'is_descriptor', 'fields', 'asdict', 'calcsize',
    'baseunits', 'byteorder', 'bitorder', 'field', 'Field', 'is_field',
    'field_descriptors',
    'BinFieldDescriptor', 'get_field_descriptor', 'set_field_descriptor',
    'BASEUNITS_ATTR_NAME', 'BYTEORDER_ATTR_NAME', 'BITORDER_ATTR_NAME',
    'METADATA_KEY',
]


BASEUNITS_ATTR_NAME = '__bpack_baseunits__'
BYTEORDER_ATTR_NAME = '__bpack_byteorder__'
BITORDER_ATTR_NAME = '__bpack_bitorder__'
SIZE_ATTR_NAME = '__bpack_size__'
METADATA_KEY = '__bpack_metadata__'


class DescriptorConsistencyError(ValueError):
    pass


class NotFieldDescriptorError(TypeError):
    pass


def _resolve_type(type_):
    """Remove type annotations.

    Replace :class:`typing.Annotated` types with the corresponding
    not-annotated ones.
    """
    if bpack.utils.is_sequence_type(type_):
        etype = bpack.utils.effective_type(type_)
        rtype = copy.copy(type_)
        rtype.__args__ = (etype,)
    elif bpack.typing.is_annotated(type_):
        rtype = bpack.utils.effective_type(type_)
    else:
        rtype = type_
    return rtype


@dataclasses.dataclass
class BinFieldDescriptor:
    """Descriptor for bpack fields."""

    type: Optional[Type] = None
    size: Optional[int] = None          # item size
    offset: Optional[int] = None
    signed: Optional[bool] = None
    repeat: Optional[int] = None        # number of items
    # converter: Optional[Callable] = None

    def _validate_type(self):
        if self.type is None:
            raise TypeError(f'invalid type "{self.type}"')

    def _validate_size(self):
        msg = f'invalid size: {self.size!r} (must be a positive integer)'
        if not isinstance(self.size, int):
            raise TypeError(msg)
        if self.size <= 0:
            raise ValueError(msg)

    def _validate_offset(self):
        msg = f'invalid offset: {self.offset!r} (must be an integer >= 0)'
        if not isinstance(self.offset, int):
            raise TypeError(msg)
        if self.offset < 0:
            raise ValueError(msg)

    def _validate_signed(self):
        if not isinstance(self.signed, bool):
            raise TypeError(
                f'invalid "signed" parameter: {self.signed!r} '
                f'(must be a bool or None)')

    def _validate_repeat(self):
        msg = f'invalid repeat: {self.repeat!r} (must be a positive)'
        if not isinstance(self.repeat, int):
            raise TypeError(msg)
        if self.repeat < 1:
            raise ValueError(msg)

    def _validate_enum_type(self):
        assert issubclass(self.type, enum.Enum)
        # perform checks on supported enum.Enum types
        bpack.utils.enum_item_type(self.type)

    def __post_init__(self):
        """Finalize BinFieldDescriptor instance initialization."""
        if self.offset is not None:
            self._validate_offset()

        if self.size is not None:
            self._validate_size()

        if self.signed is not None:
            self._validate_signed()

        if self.repeat is not None:
            self._validate_repeat()

    def validate(self):
        """Perform validity check on the BinFieldDescriptor instance."""
        self._validate_type()
        self._validate_size()
        if self.offset is not None:
            self._validate_offset()
        if self.signed is not None:
            self._validate_signed()
            if not self.is_int_type():
                warnings.warn(
                    f'the "signed" parameter will be ignored for non-integer '
                    f'type: "{self.type}"')
        if self.repeat is not None:
            self._validate_repeat()
            if not self.is_sequence_type() and self.repeat is not None:
                raise TypeError(
                    f'repeat parameter specified for non-sequence type: '
                    f'{self.type}')
        if bpack.utils.is_enum_type(self.type):
            self._validate_enum_type()
        elif self.is_sequence_type() and self.repeat is None:
            raise TypeError(
                f'no "repeat" parameter specified for sequence type '
                f'{self.type}')

    def is_int_type(self) -> bool:
        return bpack.utils.is_int_type(self.type)

    def is_sequence_type(self) -> bool:
        return bpack.utils.is_sequence_type(self.type, error=True)

    def is_enum_type(self) -> bool:
        return bpack.utils.is_enum_type(self.type)

    @property
    def total_size(self):
        """Total size in bytes of the field (considering all item)."""
        repeat = self.repeat if self.repeat is not None else 1
        return self.size * repeat

    @staticmethod
    def _is_compatible_param(old, new):
        if old is not None and new is not None and old != new:
            return False
        return True

    def update_from_type(self, type_: Type):
        if self.type is not None:
            raise TypeError('the type attribute is already set')
        if bpack.typing.is_annotated(type_):
            _, params = bpack.typing.get_args(type_)
            valid = True
            if not self._is_compatible_param(self.size, params.size):
                valid = False
            if not self._is_compatible_param(self.signed, params.signed):
                valid = False
            if not valid:
                raise DescriptorConsistencyError(
                    f'type string "{params}" is incompatible with the '
                    f'field descriptor {self}')

            self.type = params.type
            if self.signed is None:
                self.signed = params.signed
            if self.size is None:
                self.size = params.size
        elif bpack.utils.is_sequence_type(type_):
            etype = bpack.utils.effective_type(type_, keep_annotations=True)
            self.update_from_type(etype)
            self.type = _resolve_type(type_)
        else:
            self.type = type_


Field = dataclasses.Field


def field(*, size: Optional[int] = None, offset: Optional[int] = None,
          signed: Optional[bool] = None, repeat: Optional[int] = None,
          metadata=None, **kwargs) -> Field:
    """Initialize a field descriptor.

    Returned object is a :class:`Field` instance with metadata properly
    initialized to describe the field of a binary record.
    """
    field_descr = BinFieldDescriptor(size=size, offset=offset, signed=signed,
                                     repeat=repeat)
    metadata = metadata.copy() if metadata is not None else {}
    metadata[METADATA_KEY] = types.MappingProxyType(
        dataclasses.asdict(field_descr))
    return dataclasses.field(metadata=metadata, **kwargs)


def is_field(obj) -> bool:
    """Return true if an ``obj`` can be considered is a field descriptor."""
    return (isinstance(obj, Field) and obj.metadata and
            METADATA_KEY in obj.metadata)


def _update_field_metadata(field_, **kwargs):
    metadata = field_.metadata.copy() if field_.metadata is not None else {}
    metadata.update(**kwargs)
    field_.metadata = types.MappingProxyType(metadata)
    return field_


def get_field_descriptor(field: Field,                                  # noqa
                         validate: bool = True) -> BinFieldDescriptor:
    """Return the field descriptor attached to a :class:`Field`."""
    if not is_field(field):
        raise NotFieldDescriptorError(f'not a field descriptor: {field}')
    field_descr = BinFieldDescriptor(**field.metadata[METADATA_KEY])
    field_descr.update_from_type(field.type)
    if validate:
        field_descr.validate()
    return field_descr


def set_field_descriptor(field: Field, descriptor: BinFieldDescriptor,  # noqa
                         validate: bool = True) -> Field:
    """Set the field metadata according to the specified descriptor."""
    if validate:
        descriptor.validate()
    field_descr_metadata = {
        k: v for k, v in dataclasses.asdict(descriptor).items()
        if v is not None
    }
    type_ = field_descr_metadata.pop('type', None)
    if bpack.typing.is_annotated(field.type):
        field_type, _ = bpack.typing.get_args(field.type)
    else:
        field_type = field.type
    if type_ != _resolve_type(field_type):
        raise TypeError(
            f'type mismatch between BinFieldDescriptor.type ({type_!r}) and '
            f'filed.type ({field.type!r})')
    new_metadata = {
        METADATA_KEY: types.MappingProxyType(field_descr_metadata),
    }
    _update_field_metadata(field, **new_metadata)
    return field


_DEFAULT_SIZE_MAP = {
    EBaseUnits.BYTES: {
        bool: 1,
        # int: 4,
        # float: 4,
    },
    EBaseUnits.BITS: {
        bool: 1,
        # int: 32,
        # float: 32,
    },
}


def _get_default_size(type_,
                      baseunits: EBaseUnits) -> Union[int, None]:       # noqa
    if is_descriptor(type_):
        return calcsize(type_, baseunits)

    etype = bpack.utils.effective_type(type_)

    # if bpack.utils.is_enum_type(type_):
    #     if bpack.utils.is_int_type(type_):
    #         signbit = 1 if any(item.value < 0 for item in type_) else 0
    #         bits = signbit + max(item.value.bit_lenght() for item in type_)
    #         if baseunits is EBaseUnits.BITS:
    #             if bits <= 8:
    #                 return 1
    #             elif bits <= 16:
    #                 return 2
    #             elif bits <= 32:
    #                 return 4
    #             else:
    #                 return 8
    #         else:
    #             return bits
    #     elif issubclass(etype, str):
    #         length = max(len(item.value.encode('utf-8')) for item in type_)
    #         return length * bytes_to_baseunits
    #     elif issubclass(etype, bytes):
    #         length = max(len(item.value) for item in type_)
    #         return length * bytes_to_baseunits
    #     else:
    #         return None

    return _DEFAULT_SIZE_MAP[baseunits].get(etype)                      # noqa


def _get_effective_byteorder(byteorder: EByteOrder,                     # noqa
                             baseunits: EBaseUnits) -> EByteOrder:      # noqa
    byteorder = EByteOrder(byteorder)                                   # noqa
    effective_byteorder = byteorder
    if baseunits is EBaseUnits.BYTES:
        if byteorder in {EByteOrder.NATIVE, EByteOrder.DEFAULT}:
            effective_byteorder = EByteOrder.get_native()
    else:
        if byteorder is EByteOrder.DEFAULT:
            effective_byteorder = EByteOrder.BE
        elif byteorder in EByteOrder.NATIVE:
            effective_byteorder = EByteOrder.get_native()
    return effective_byteorder


@classdecorator
def descriptor(cls, *, size: Optional[int] = None,
               byteorder: Union[str, EByteOrder] = EByteOrder.DEFAULT,  # noqa
               bitorder: Optional[Union[str, EBitOrder]] = None,        # noqa
               baseunits: EBaseUnits = EBaseUnits.BYTES,                # noqa
               **kwargs):
    """Class decorator to define descriptors for binary records.

    It converts a dataclass into a descriptor object for binary records.

    * ensures that all fields are :class:`bpack.descriptor.Field` descriptors
    * offsets are automatically computed if necessary
    * consistency checks on offsets and sizes are performed

    :param cls:
        class to be decorated
    :param size:
        the size (expressed in *base units*) of the binary record
    :param byteorder:
        the byte-order of the binary record
    :param bitorder:
        the bit-order of the binary record
        (must be ``None`` if the *base units* are bytes)
    :param baseunits:
        the base units (:data:`bpack.enums.EBaseUnits.BITS` or
        :data:`bpack.enums.EBaseUnits.BYTES`) used to specify the
        binary record descriptor

    It is also possible to specify as additional keyword arguments all the
    parameters accepted by :func:`dataclasses.dataclass`.
    """
    baseunits = EBaseUnits(baseunits)                                   # noqa
    byteorder = EByteOrder(byteorder)                                   # noqa

    if dataclasses.is_dataclass(cls):
        warnings.warn('the explicit use of dataclasses is deprecated',
                      category=DeprecationWarning)
    else:
        cls = dataclasses.dataclass(cls, **kwargs)

    fields_ = dataclasses.fields(cls)

    # Initialize to a dummy value with initial offset + size = 0
    prev_field_descr = BinFieldDescriptor(size=None, offset=0)
    prev_field_descr.size = 0  # trick to bypass checks on BinFieldDescriptor
    content_size = 0

    for idx, field_ in enumerate(fields_):
        assert isinstance(field_, Field)

        # NOTE: this is ensured by dataclasses but not by attr
        assert field_.type is not None

        if bpack.typing.is_annotated(field_.type):
            # check byteorder
            _, params = bpack.typing.get_args(field_.type)
            if params.byteorder:
                effective_byteorder = _get_effective_byteorder(byteorder,
                                                               baseunits)
                if params.byteorder != effective_byteorder:
                    raise DescriptorConsistencyError(
                        f'the byteorder of field {field_.name} '
                        f'("{params.byteorder}" is not consistent with the '
                        f'descriptor byteorder ({byteorder}) )')

        try:
            field_descr = get_field_descriptor(field_, validate=False)
        except NotFieldDescriptorError:
            field_descr = BinFieldDescriptor()
            if isinstance(field_, Field):
                field_descr.update_from_type(field_.type)

        if field_descr.size is None:
            field_descr.size = _get_default_size(field_descr.type, baseunits)

        if field_descr.size is None:
            raise TypeError(f'size not specified for field: "{field_.name}"')

        if (is_descriptor(field_descr.type) and
                calcsize(field_descr.type, baseunits) != field_descr.size):
            raise DescriptorConsistencyError(
                f'mismatch between field.size ({field_descr.size}) and size '
                f'of field.type ({calcsize(field_descr.type, baseunits)}) '
                f'in field "{field_.name}"')

        auto_offset = prev_field_descr.offset + prev_field_descr.total_size

        if field_descr.offset is None:
            field_descr.offset = auto_offset
        elif field_descr.offset < auto_offset:
            raise DescriptorConsistencyError(
                f'invalid offset for filed n. {idx}: {field_}')

        set_field_descriptor(field_, field_descr)
        prev_field_descr = field_descr
        content_size += field_descr.size

    field_descr = get_field_descriptor(fields_[-1])
    auto_size = field_descr.offset + field_descr.total_size
    assert auto_size >= content_size  # this should be already checked above

    if size is None:
        size = auto_size
    elif int(size) != size:
        raise TypeError(f'invalid size: {size!r}')
    elif size < auto_size:
        raise DescriptorConsistencyError(
            f'the specified size ({size}) is smaller than total size of '
            f'fields ({auto_size})')

    if baseunits is EBaseUnits.BITS:
        if size % 8 != 0:
            warnings.warn('bit struct not aligned to bytes')

    if baseunits is not EBaseUnits.BITS and bitorder is not None:
        raise ValueError(
            'it is not possible to specify the "bitorder" '
            'if "baseunits" is not "BITS"')
    elif baseunits is EBaseUnits.BITS and bitorder is None:
        bitorder = EBitOrder.DEFAULT                                    # noqa

    setattr(cls, BASEUNITS_ATTR_NAME, baseunits)
    setattr(cls, BYTEORDER_ATTR_NAME, byteorder)
    setattr(cls, BITORDER_ATTR_NAME,
            EBitOrder(bitorder) if bitorder is not None else None)
    setattr(cls, SIZE_ATTR_NAME, size)

    return cls


def fields(obj) -> Sequence[Field]:
    """Return a tuple describing the fields of this descriptor."""
    return dataclasses.fields(obj)                                      # noqa


def is_descriptor(obj) -> bool:
    """Return true if ``obj`` is a descriptor or a descriptor instance."""
    try:
        return hasattr(obj, BASEUNITS_ATTR_NAME) and is_field(fields(obj)[0])
    except (TypeError, ValueError):
        # dataclass.fields(...) --> TypeError
        # attr.fields(...)      --> NotAnAttrsClassError(ValueError)
        return False
    except IndexError:
        # no fields
        return False


def asdict(obj, *, dict_factory=dict) -> dict:
    """Return the fields of a record as a new dictionary.

     The returned dictionary maps field names to field values.

    If given, 'dict_factory' will be used instead of built-in dict.
    The function applies recursively to field values that are
    dataclass instances. This will also look into built-in containers:
    tuples, lists, and dicts.
    """
    return dataclasses.asdict(obj, dict_factory=dict_factory)


def calcsize(obj, units: Optional[EBaseUnits] = None) -> int:
    """Return the size of the ``obj`` record.

    If the *units* parameter is not specified (default) then the returned
    *size* is expressed in the same *base units* of the descriptor.
    """
    if not is_descriptor(obj):
        raise TypeError(f'{obj!r} is not a descriptor')

    size = getattr(obj, SIZE_ATTR_NAME)
    baseunits_ = getattr(obj, BASEUNITS_ATTR_NAME)

    if units and units is not baseunits_:
        if units is EBaseUnits.BYTES:
            # baseunits is BITS and units is BYTES
            size = math.ceil(size / 8)
        else:
            # baseunits is BYTES and units is BITS
            size *= 8

    return size


def baseunits(obj) -> EBaseUnits:
    """Return the base units of a binary record descriptor."""
    try:
        return getattr(obj, BASEUNITS_ATTR_NAME)
    except AttributeError:
        raise TypeError(f'"{obj}" is not a descriptor')


def byteorder(obj) -> EByteOrder:
    """Return the byte order of a binary record descriptor (endianess)."""
    try:
        return getattr(obj, BYTEORDER_ATTR_NAME)
    except AttributeError:
        raise TypeError(f'"{obj}" is not a descriptor')


def bitorder(obj) -> Union[EBitOrder, None]:
    """Return the bit order of a binary record descriptor."""
    try:
        return getattr(obj, BITORDER_ATTR_NAME)
    except AttributeError:
        raise TypeError(f'"{obj}" is not a descriptor')


def field_descriptors(descriptor,                                       # noqa
                      pad: bool = False) -> Iterable[BinFieldDescriptor]:
    """Return the list of field descriptors for the input record descriptor.

    Items are instances of the :class:`BinFieldDescriptor` class describing
    characteristics of each field of the input binary record descriptor.

    If the ``pad`` parameter is set to True then also generate dummy field
    descriptors for padding elements necessary to take into account offsets
    between fields.
    """
    if pad:
        offset = 0
        for field_ in fields(descriptor):
            field_descr = get_field_descriptor(field_)
            assert field_descr.offset >= offset
            if field_descr.offset > offset:
                # padding
                yield BinFieldDescriptor(size=field_descr.offset - offset,
                                         offset=offset)
                # offset = field_.offset
            yield field_descr
            offset = field_descr.offset + field_descr.total_size

        size = calcsize(descriptor)
        if offset < size:
            # padding
            yield BinFieldDescriptor(size=size - offset, offset=offset)
    else:
        for field_ in fields(descriptor):
            yield get_field_descriptor(field_)



"""Struct based codec for binary data structures."""

import struct

from typing import Optional

import bpack
import bpack.utils
import bpack.codecs

from .enums import EBaseUnits
from .codecs import get_sequence_groups, has_codec, get_codec
from .descriptors import field_descriptors


__all__ = ['Decoder', 'decoder', 'BACKEND_NAME', 'BACKEND_TYPE']


BACKEND_NAME = 'bitstruct'
BACKEND_TYPE = EBaseUnits.BYTES


_TYPE_SIGNED_AND_SIZE_TO_STR = {
    # type, signed, size
    (bool, None, 1): 'B',
    (int, None, 1): 'b',        # default
    (int, False, 1): 'B',
    (int, True, 1): 'b',
    (int, None, 2): 'h',        # default
    (int, False, 2): 'H',
    (int, True, 2): 'h',
    (int, None, 4): 'i',        # default
    (int, False, 4): 'I',
    (int, True, 4): 'i',
    (int, None, 8): 'q',        # default
    (int, False, 8): 'Q',
    (int, True, 8): 'q',
    (float, None, 2): 'e',
    (float, None, 4): 'f',
    (float, None, 8): 'd',
    (bytes, None, None): 's',
    (str, None, None): 's',
    (None, None, None): 'x',        # padding
}


_DEFAULT_SIZE = {
    bool: 1,
    int: 4,
    float: 4,
    bytes: 1,
    # str: 1,
}


def _to_fmt(type_, size: Optional[int] = None, order: str = '',
            signed: Optional[bool] = None,
            repeat: Optional[int] = None) -> str:
    size = _DEFAULT_SIZE.get(type_, size) if size is None else size

    assert size is not None and size > 0
    assert order in ('', '>', '<', '=', '@', '!'), f'invalid order: {order!r}'
    assert signed in (True, False, None)

    if has_codec(type_, bpack.codecs.Decoder):
        decoder_ = get_codec(type_)
        if isinstance(decoder_, Decoder):
            return decoder_.format
    elif (bpack.is_descriptor(type_) and
          bpack.baseunits(type_) is Decoder.baseunits):
        decoder_ = Decoder(type_)
        return decoder_.format

    etype = bpack.utils.effective_type(type_)
    repeat = 1 if repeat is None else repeat
    try:
        if etype in (str, bytes, None):  # none is for padding bytes
            key = (etype, signed, None)
            return f'{order}{size}{_TYPE_SIGNED_AND_SIZE_TO_STR[key]}' * repeat
        else:
            key = (etype, signed, size)
            return f'{order}{repeat}{_TYPE_SIGNED_AND_SIZE_TO_STR[key]}'
    except KeyError:
        raise TypeError(
            f'unable to generate format string for '
            f'type="{type_}", size="{size}", order="{order}", '
            f'signed="{signed}", repeat="{repeat}"')


def _enum_converter_factory(type_, converters=None):
    converters = converters if converters is not None else dict()
    enum_item_type = bpack.utils.enum_item_type(type_)
    if enum_item_type in converters:
        base_converter = converters[enum_item_type]
        return lambda x: type_(base_converter(x))
    else:
        return type_


class Decoder(bpack.codecs.Decoder):
    """Struct based data decoder.

    Default byte-order: MSB.
    """

    baseunits = EBaseUnits.BYTES

    def __init__(self, descriptor):
        """Initializer.

        The *descriptor* parameter* is a bpack record descriptor.
        """
        super().__init__(descriptor)

        assert bpack.bitorder(descriptor) is None

        byteorder = bpack.byteorder(descriptor)
        # assert all(descr.order for descr in field_descriptors(descriptor))
        byteorder = byteorder.value

        # NOTE: struct expects that the byteorder specifier is used only
        #       once at the beginning of the format string
        fmt = byteorder + ''.join(
            _to_fmt(field_descr.type, field_descr.size, order='',
                    repeat=field_descr.repeat)
            for field_descr in field_descriptors(descriptor, pad=True)
        )

        converters_map = {
            str: lambda s: s.decode('ascii'),
            bool: bool,
        }
        converters_map.update(
            (field_descr.type,
             _enum_converter_factory(field_descr.type, converters_map))
            for field_descr in field_descriptors(descriptor)
            if bpack.utils.is_enum_type(field_descr.type)
        )

        self._codec = struct.Struct(fmt)
        self._converters = [
            (idx, converters_map[field_descr.type])
            for idx, field_descr in enumerate(
                field_descriptors(self.descriptor))
            if field_descr.type in converters_map
        ]
        self._groups = get_sequence_groups(descriptor)

    @property
    def format(self) -> str:
        """Return the :mod:`struct` format string."""
        return self._codec.format

    def decode(self, data: bytes):
        """Decode binary data and return a record object."""
        values = list(self._codec.unpack(data))

        for idx, func in self._converters:
            values[idx] = func(values[idx])

        for type_, slice_ in self._groups[::-1]:
            subtype = type_(values[slice_])
            del values[slice_]
            values.insert(slice_.start, subtype)

        return self.descriptor(*values)


decoder = bpack.codecs.make_codec_decorator(Decoder)



#!/usr/bin/env python3

import os
import setuptools


ROOT = os.path.dirname(__file__)


def get_version(filename, strip_extra=False):
    import re
    from distutils.version import LooseVersion
    # import packaging.version

    with open(filename) as fd:
        data = fd.read()

    mobj = re.search(
        r'''^__version__\s*=\s*(?P<quote>['"])(?P<version>.*)(?P=quote)''',
        data, re.MULTILINE)
    version = LooseVersion(mobj.group('version'))

    if strip_extra:
        return '.'.join(map(str, version.version[:3]))
    else:
        return version.vstring


cfg = dict(
    version=get_version(os.path.join(ROOT, 'bpack', '__init__.py')),
)


if __name__ == '__main__':
    setuptools.setup(**cfg)



"""Test bpack decoders."""

import sys
import enum

from typing import List, Sequence

import pytest

import bpack
import bpack.st
import bpack.codecs

try:
    import bpack.bs as bpack_bs
except ImportError:                                         # pragma: no cover
    bpack_bs = None

try:
    import bpack.ba as bpack_ba
except ImportError:                                         # pragma: no cover
    bpack_ba = None

try:
    import bpack.np as bpack_np
except ImportError:                                         # pragma: no cover
    bpack_np = None


skipif = pytest.mark.skipif

BITS_BACKENDS = [
    pytest.param(bpack_bs, id='bs',
                 marks=skipif(not bpack_bs, reason='not available')),
    pytest.param(bpack_ba, id='ba',
                 marks=skipif(not bpack_ba, reason='not available')),
]
BYTES_BACKENDS = [
    pytest.param(bpack.st, id='st'),
    pytest.param(bpack_np, id='np',
                 marks=skipif(not bpack_np, reason='not available')),
]
ALL_BACKENDS = BITS_BACKENDS + BYTES_BACKENDS


@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_backend(backend):
    assert hasattr(backend, 'BACKEND_NAME')
    assert hasattr(backend, 'BACKEND_TYPE')


@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_attrs(backend):
    @backend.decoder
    @bpack.descriptor(baseunits=backend.BACKEND_TYPE)
    class Record:
        field_1: int = bpack.field(size=4, default=0)
        field_2: int = bpack.field(size=4, default=1)

    assert hasattr(Record, bpack.descriptors.BASEUNITS_ATTR_NAME)
    assert hasattr(Record, bpack.descriptors.BYTEORDER_ATTR_NAME)
    assert hasattr(Record, bpack.descriptors.BITORDER_ATTR_NAME)
    assert hasattr(Record, bpack.descriptors.SIZE_ATTR_NAME)
    assert hasattr(Record, bpack.codecs.CODEC_ATTR_NAME)


@bpack.descriptor(baseunits=bpack.EBaseUnits.BITS,
                  byteorder=bpack.EByteOrder.BE,
                  bitorder=bpack.EBitOrder.MSB,
                  frozen=True)
class BitRecordBeMsb:
    # default (unsigned)
    field_01: bool = bpack.field(size=1, default=True)
    field_02: int = bpack.field(size=3, default=4)
    field_03: int = bpack.field(size=12, default=2048)
    field_04: float = bpack.field(size=32, default=1.)
    field_05: bytes = bpack.field(size=24, default=b'abc')
    field_06: str = bpack.field(size=24, default='ABC')
    # 4 padding bits ([96:100])
    field_08: int = bpack.field(size=28, default=134217727, offset=100)

    # signed
    field_11: bool = bpack.field(size=1, default=False)
    field_12: int = bpack.field(size=3, default=-4, signed=True)
    field_13: int = bpack.field(size=12, default=-2048, signed=True)
    field_18: int = bpack.field(size=32, default=-2**31, signed=True)

    # unsigned
    field_21: bool = bpack.field(size=1, default=True)
    field_22: int = bpack.field(size=3, default=4, signed=False)
    field_23: int = bpack.field(size=12, default=2048, signed=False)
    field_28: int = bpack.field(size=32, default=2**31, signed=False)


BIT_ENCODED_DATA_BE_MSB = b''.join([
    # default (unsigned)
    bytes([
        0b11001000, 0b00000000,                             # fields 1 to 3
        0b00111111, 0b10000000, 0b00000000, 0b00000000,     # field_4 (float32)
    ]),
    b'abc',                                                 # field_5 (bytes)
    'ABC'.encode('ascii'),                                  # field_6 (str)
    bytes([                                         # 4 padding bits + field_8
        0b00000111, 0b11111111, 0b11111111, 0b11111111,
    ]),

    # signed
    bytes([
        0b01001000, 0b00000000,                             # fields 11 to 13
        0b10000000, 0b00000000, 0b00000000, 0b00000000,     # field_18 (uint32)
    ]),

    # unsigned
    bytes([
        0b11001000, 0b00000000,                             # fields 21 to 23
        0b10000000, 0b00000000, 0b00000000, 0b00000000,     # field_28 (sint32)
    ]),
])


@bpack.descriptor(baseunits=bpack.EBaseUnits.BITS,
                  byteorder=bpack.EByteOrder.LE,
                  bitorder=bpack.EBitOrder.MSB,
                  frozen=True)
class BitRecordLeMsb:
    # default (unsigned)
    field_01: bool = bpack.field(size=1, default=True)
    field_02: int = bpack.field(size=3, default=4)
    field_03: int = bpack.field(size=12, default=2048)
    field_04: float = bpack.field(size=32, default=1.)
    field_05: bytes = bpack.field(size=24, default=b'abc')
    field_06: str = bpack.field(size=24, default='ABC')
    # 4 padding bits ([96:100])
    field_08: int = bpack.field(size=28, default=134217727, offset=100)

    # signed
    field_11: bool = bpack.field(size=1, default=False)
    field_12: int = bpack.field(size=3, default=-4, signed=True)
    field_13: int = bpack.field(size=12, default=-2048, signed=True)
    field_18: int = bpack.field(size=32, default=-2**31, signed=True)

    # unsigned
    field_21: bool = bpack.field(size=1, default=True)
    field_22: int = bpack.field(size=3, default=4, signed=False)
    field_23: int = bpack.field(size=12, default=2048, signed=False)
    field_28: int = bpack.field(size=32, default=2**31, signed=False)


BIT_ENCODED_DATA_LE_MSB = b''.join([
    # default (unsigned)
    bytes([
        0b11000000, 0b10000000,                             # fields 1 to 3
        0b00000000, 0b00000000, 0b10000000, 0b00111111,     # field_4 (float32)
    ]),
    b'abc',                                                 # field_5 (bytes)
    'ABC'.encode('ascii'),                                  # field_6 (str)
    bytes([                                         # 4 padding bits + field_8
        0b00001111, 0b11111111, 0b11111111, 0b01111111,
    ]),

    # signed
    bytes([
        0b01000000, 0b10000000,                             # fields 11 to 13
        0b00000000, 0b00000000, 0b00000000, 0b10000000,     # field_18 (sint32)
    ]),

    # unsigned
    bytes([
        0b11000000, 0b10000000,                             # fields 21 to 23
        0b00000000, 0b00000000, 0b00000000, 0b10000000,     # field_28 (uint32)
    ]),
])


@bpack.descriptor(baseunits=bpack.EBaseUnits.BITS,
                  byteorder=bpack.EByteOrder.BE,
                  bitorder=bpack.EBitOrder.LSB,
                  frozen=True)
class BitRecordBeLsb:
    # default (unsigned)
    field_01: bool = bpack.field(size=1, default=True)
    field_02: int = bpack.field(size=3, default=4)
    field_03: int = bpack.field(size=12, default=2048)
    field_04: float = bpack.field(size=32, default=1.)
    field_05: bytes = bpack.field(size=24, default=b'abc')
    field_06: str = bpack.field(size=24, default='ABC')
    # 4 padding bits ([96:100])
    field_08: int = bpack.field(size=28, default=134217727, offset=100)

    # signed
    field_11: bool = bpack.field(size=1, default=False)
    field_12: int = bpack.field(size=3, default=-4, signed=True)
    field_13: int = bpack.field(size=12, default=-2048, signed=True)
    field_18: int = bpack.field(size=32, default=-2**31, signed=True)

    # unsigned
    field_21: bool = bpack.field(size=1, default=True)
    field_22: int = bpack.field(size=3, default=4, signed=False)
    field_23: int = bpack.field(size=12, default=2048, signed=False)
    field_28: int = bpack.field(size=32, default=2**31, signed=False)


BIT_ENCODED_DATA_BE_LSB = b''.join([
    # default (unsigned)
    bytes([
        0b10010000, 0b00000001,                             # fields 1 to 3
        0b00000000, 0b00000000, 0b00000001, 0b11111100,     # field_4 (float32)
    ]),
    bytes([0b11000110, 0b01000110, 0b10000110]),    # field_5 (bytes) - b'abc'
    bytes([0b11000010, 0b01000010, 0b10000010]),    # field_6 (str)   - 'ABC'
    bytes([                                         # 4 padding bits + field_8
        0b00001111, 0b11111111, 0b11111111, 0b11111110,
    ]),

    # signed
    bytes([
        0b00010000, 0b00000001,                             # fields 11 to 13
        0b00000000, 0b00000000, 0b00000000, 0b00000001,     # field_18 (uint32)
    ]),

    # unsigned
    bytes([
        0b10010000, 0b00000001,                             # fields 21 to 23
        0b00000000, 0b00000000, 0b00000000, 0b00000001,     # field_28 (sint32)
    ]),
])


@bpack.descriptor(baseunits=bpack.EBaseUnits.BITS,
                  byteorder=bpack.EByteOrder.LE,
                  bitorder=bpack.EBitOrder.LSB,
                  frozen=True)
class BitRecordLeLsb:
    # default (unsigned)
    field_01: bool = bpack.field(size=1, default=True)
    field_02: int = bpack.field(size=3, default=4)
    field_03: int = bpack.field(size=12, default=2048)
    field_04: float = bpack.field(size=32, default=1.)
    field_05: bytes = bpack.field(size=24, default=b'abc')
    field_06: str = bpack.field(size=24, default='ABC')
    # 4 padding bits ([96:100])
    field_08: int = bpack.field(size=28, default=134217727, offset=100)

    # signed
    field_11: bool = bpack.field(size=1, default=False)
    field_12: int = bpack.field(size=3, default=-4, signed=True)
    field_13: int = bpack.field(size=12, default=-2048, signed=True)
    field_18: int = bpack.field(size=32, default=-2**31, signed=True)

    # unsigned
    field_21: bool = bpack.field(size=1, default=True)
    field_22: int = bpack.field(size=3, default=4, signed=False)
    field_23: int = bpack.field(size=12, default=2048, signed=False)
    field_28: int = bpack.field(size=32, default=2**31, signed=False)


BIT_ENCODED_DATA_LE_LSB = b''.join([
    # default (unsigned)
    bytes([
        0b10010001, 0b00000000,                             # fields 1 to 3
        0b11111100, 0b00000001, 0b00000000, 0b00000000,     # field_4 (float32)
    ]),
    bytes([0b11000110, 0b01000110, 0b10000110]),    # field_5 (bytes) - b'abc'
    bytes([0b11000010, 0b01000010, 0b10000010]),    # field_6 (str)   - 'ABC'
    bytes([                                         # 4 padding bits + field_8
        0b00001110, 0b11111111, 0b11111111, 0b11111111,
    ]),

    # signed
    bytes([
        0b00010001, 0b00000000,                             # fields 11 to 13
        0b00000001, 0b00000000, 0b00000000, 0b00000000,     # field_18 (sint32)
    ]),

    # unsigned
    bytes([
        0b10010001, 0b00000000,                             # fields 21 to 23
        0b00000001, 0b00000000, 0b00000000, 0b00000000,     # field_28 (uint32)
    ]),
])


@bpack.descriptor(baseunits=bpack.EBaseUnits.BYTES,
                  byteorder=bpack.EByteOrder.BE,
                  frozen=True)
class ByteRecordBe:
    field_01: bool = bpack.field(size=1, default=False)

    field_02: int = bpack.field(size=1, default=1)
    field_03: int = bpack.field(size=1, default=-1, signed=True)
    field_04: int = bpack.field(size=1, default=+1, signed=False)

    field_05: int = bpack.field(size=2, default=2)
    field_06: int = bpack.field(size=2, default=-2, signed=True)
    field_07: int = bpack.field(size=2, default=+2, signed=False)

    field_08: int = bpack.field(size=4, default=4)
    field_09: int = bpack.field(size=4, default=-4, signed=True)
    field_10: int = bpack.field(size=4, default=+4, signed=False)

    field_11: int = bpack.field(size=8, default=8)
    field_12: int = bpack.field(size=8, default=-8, signed=True)
    field_13: int = bpack.field(size=8, default=+8, signed=False)

    field_14: float = bpack.field(size=2, default=10.)
    field_15: float = bpack.field(size=4, default=100.)
    field_16: float = bpack.field(size=8, default=1000.)

    field_17: bytes = bpack.field(size=3, default=b'abc')
    field_18: str = bpack.field(size=3, default='ABC')

    # 4 padding bytes ([66:70]) b'xxxx'

    field_20: bytes = bpack.field(size=4, offset=70, default=b'1234')


BYTE_ENCODED_DATA_BE = bytes([
    False,                                              # bool (8 bit)

    0b00000001,                                         # +1    sint8
    0b11111111,                                         # -1    sint8
    0b00000001,                                         # +1    uint8

    0b00000000, 0b00000010,                             # +2    sint16
    0b11111111, 0b11111110,                             # -2    sint16
    0b00000000, 0b00000010,                             # +2    uint16

    0b00000000, 0b00000000, 0b00000000, 0b00000100,     # +4    sint32
    0b11111111, 0b11111111, 0b11111111, 0b11111100,     # -4    sint32
    0b00000000, 0b00000000, 0b00000000, 0b00000100,     # +4    uint32

    0b00000000, 0b00000000, 0b00000000, 0b00000000,
    0b00000000, 0b00000000, 0b00000000, 0b00001000,     # +8    sint64
    0b11111111, 0b11111111, 0b11111111, 0b11111111,
    0b11111111, 0b11111111, 0b11111111, 0b11111000,     # -8    sint64
    0b00000000, 0b00000000, 0b00000000, 0b00000000,
    0b00000000, 0b00000000, 0b00000000, 0b00001000,     # +8    uint64

    0b01001001, 0b00000000,                             # 10    float16
    0b01000010, 0b11001000, 0b00000000, 0b00000000,     # 100   float32
    0b01000000, 0b10001111, 0b01000000, 0b00000000,
    0b00000000, 0b00000000, 0b00000000, 0b00000000,     # 1000  float64

    0b01100001, 0b01100010, 0b01100011,                 # b'abc'    bytes
    0b01000001, 0b01000010, 0b01000011,                 # 'ABC'     str

    # 4 padding bytes ([36:40])
    0b01111000, 0b01111000, 0b01111000, 0b01111000,     # b'xxxx'   bytes

    # field_20
    0b00110001, 0b00110010, 0b00110011, 0b00110100,     # b'1234'   bytes
])


@bpack.descriptor(baseunits=bpack.EBaseUnits.BYTES,
                  byteorder=bpack.EByteOrder.LE,
                  frozen=True)
class ByteRecordLe:
    field_01: bool = bpack.field(size=1, default=False)

    field_02: int = bpack.field(size=1, default=1)
    field_03: int = bpack.field(size=1, default=-1, signed=True)
    field_04: int = bpack.field(size=1, default=+1, signed=False)

    field_05: int = bpack.field(size=2, default=2)
    field_06: int = bpack.field(size=2, default=-2, signed=True)
    field_07: int = bpack.field(size=2, default=+2, signed=False)

    field_08: int = bpack.field(size=4, default=4)
    field_09: int = bpack.field(size=4, default=-4, signed=True)
    field_10: int = bpack.field(size=4, default=+4, signed=False)

    field_11: int = bpack.field(size=8, default=8)
    field_12: int = bpack.field(size=8, default=-8, signed=True)
    field_13: int = bpack.field(size=8, default=+8, signed=False)

    field_14: float = bpack.field(size=2, default=10.)
    field_15: float = bpack.field(size=4, default=100.)
    field_16: float = bpack.field(size=8, default=1000.)

    field_17: bytes = bpack.field(size=3, default=b'abc')
    field_18: str = bpack.field(size=3, default='ABC')

    # 4 padding bytes ([66:70]) b'xxxx'

    field_20: bytes = bpack.field(size=4, offset=70, default=b'1234')


BYTE_ENCODED_DATA_LE = bytes([
    False,                                              # bool (8 bit)

    0b00000001,                                         # +1    sint8
    0b11111111,                                         # -1    sint8
    0b00000001,                                         # +1    uint8

    0b00000010, 0b00000000,                             # +2    sint16
    0b11111110, 0b11111111,                             # -2    sint16
    0b00000010, 0b00000000,                             # +2    uint16

    0b00000100, 0b00000000, 0b00000000, 0b00000000,     # +4    sint32
    0b11111100, 0b11111111, 0b11111111, 0b11111111,     # -4    sint32
    0b00000100, 0b00000000, 0b00000000, 0b00000000,     # +4    uint32

    0b00001000, 0b00000000, 0b00000000, 0b00000000,
    0b00000000, 0b00000000, 0b00000000, 0b00000000,     # +8    sint64
    0b11111000, 0b11111111, 0b11111111, 0b11111111,
    0b11111111, 0b11111111, 0b11111111, 0b11111111,     # -8    sint64
    0b00001000, 0b00000000, 0b00000000, 0b00000000,
    0b00000000, 0b00000000, 0b00000000, 0b00000000,     # +8    uint64

    0b00000000, 0b01001001,                             # 10    float16
    0b00000000, 0b00000000, 0b11001000, 0b01000010,     # 100   float32
    0b00000000, 0b00000000, 0b00000000, 0b00000000,
    0b00000000, 0b01000000, 0b10001111, 0b01000000,     # 1000  float64

    0b01100001, 0b01100010, 0b01100011,                 # b'abc'    bytes
    0b01000001, 0b01000010, 0b01000011,                 # 'ABC'     str

    # 4 padding bytes ([66:70])
    0b01111000, 0b01111000, 0b01111000, 0b01111000,     # b'xxxx'   bytes

    # field_20
    0b00110001, 0b00110010, 0b00110011, 0b00110100,     # b'1234'   bytes
])


@pytest.mark.parametrize(
    'backend, Record, encoded_data',
    [pytest.param(bpack.st, ByteRecordBe, BYTE_ENCODED_DATA_BE, id='st BE'),
     pytest.param(bpack.st, ByteRecordLe, BYTE_ENCODED_DATA_LE, id='st LE'),
     pytest.param(
         bpack_np, ByteRecordBe, BYTE_ENCODED_DATA_BE, id='np BE',
         marks=pytest.mark.skipif(not bpack_np, reason='not available')),
     pytest.param(
         bpack_np, ByteRecordLe, BYTE_ENCODED_DATA_LE, id='np LE',
         marks=pytest.mark.skipif(not bpack_np, reason='not available')),
     pytest.param(
         bpack_bs, BitRecordBeMsb, BIT_ENCODED_DATA_BE_MSB, id='bs BE MSB',
         marks=pytest.mark.skipif(not bpack_bs, reason='not available')),
     pytest.param(
         bpack_bs, BitRecordLeMsb, BIT_ENCODED_DATA_LE_MSB, id='bs LE MSB',
         marks=pytest.mark.skipif(not bpack_bs, reason='not available')),
     pytest.param(
         bpack_bs, BitRecordBeLsb, BIT_ENCODED_DATA_BE_LSB, id='bs BE LSB',
         marks=pytest.mark.skipif(not bpack_bs, reason='not available')),
     pytest.param(
         bpack_bs, BitRecordLeLsb, BIT_ENCODED_DATA_LE_LSB, id='bs LE LSB',
         marks=pytest.mark.skipif(not bpack_bs, reason='not available')),
     pytest.param(
         bpack_ba, BitRecordBeMsb, BIT_ENCODED_DATA_BE_MSB, id='ba BE MSB',
         marks=pytest.mark.skipif(not bpack_ba, reason='not available'))])
def test_decoder(backend, Record, encoded_data):                        # noqa
    decoded_data = Record()                                             # noqa
    decoder = backend.Decoder(Record)
    assert hasattr(decoder, 'baseunits')
    assert decoder.baseunits is bpack.baseunits(Record)
    record = decoder.decode(encoded_data)
    assert record == decoded_data


@pytest.mark.parametrize(
    'backend, Record, encoded_data',
    [pytest.param(bpack.st, ByteRecordBe, BYTE_ENCODED_DATA_BE, id='st BE'),
     pytest.param(bpack.st, ByteRecordLe, BYTE_ENCODED_DATA_LE, id='st LE'),
     pytest.param(
         bpack_np, ByteRecordBe, BYTE_ENCODED_DATA_BE, id='np BE',
         marks=pytest.mark.skipif(not bpack_np, reason='not available')),
     pytest.param(
         bpack_np, ByteRecordLe, BYTE_ENCODED_DATA_LE, id='np LE',
         marks=pytest.mark.skipif(not bpack_np, reason='not available')),
     pytest.param(
         bpack_bs, BitRecordBeMsb, BIT_ENCODED_DATA_BE_MSB, id='bs BE MSB',
         marks=pytest.mark.skipif(not bpack_bs, reason='not available')),
     pytest.param(
         bpack_bs, BitRecordLeMsb, BIT_ENCODED_DATA_LE_MSB, id='bs LE MSB',
         marks=pytest.mark.skipif(not bpack_bs, reason='not available')),
     pytest.param(
         bpack_bs, BitRecordBeLsb, BIT_ENCODED_DATA_BE_LSB, id='bs BE LSB',
         marks=pytest.mark.skipif(not bpack_bs, reason='not available')),
     pytest.param(
         bpack_bs, BitRecordLeLsb, BIT_ENCODED_DATA_LE_LSB, id='bs LE LSB',
         marks=pytest.mark.skipif(not bpack_bs, reason='not available')),
     pytest.param(
         bpack_ba, BitRecordBeMsb, BIT_ENCODED_DATA_BE_MSB, id='ba BE MSB',
         marks=pytest.mark.skipif(not bpack_ba, reason='not available'))])
def test_decoder_func(backend, Record, encoded_data):                   # noqa
    decoded_data = Record()                                             # noqa
    record_type = backend.decoder(Record)
    record = record_type.frombytes(encoded_data)
    assert record == decoded_data


@pytest.mark.parametrize('backend', BITS_BACKENDS)
def test_bit_decoder_decorator(backend):
    @backend.decoder
    @bpack.descriptor(baseunits=bpack.EBaseUnits.BITS, frozen=True)
    class Record:
        field_01: bool = bpack.field(size=1, default=True)
        field_02: int = bpack.field(size=3, default=4)
        field_03: int = bpack.field(size=12, default=2048)
        field_04: float = bpack.field(size=32, default=1.)
        field_05: bytes = bpack.field(size=24, default=b'abc')
        field_06: str = bpack.field(size=24, default='ABC')
        # 4 padding bits ([96:100])  0b1111
        field_08: int = bpack.field(size=28, default=134217727, offset=100)

    decoded_data = Record()
    encoded_data = BIT_ENCODED_DATA_BE_MSB[:bpack.calcsize(Record)]
    record = Record.frombytes(encoded_data)

    assert record.field_01 == decoded_data.field_01
    assert record.field_02 == decoded_data.field_02
    assert record.field_03 == decoded_data.field_03
    assert record.field_04 == decoded_data.field_04
    assert record.field_05 == decoded_data.field_05
    assert record.field_06 == decoded_data.field_06
    assert record.field_08 == decoded_data.field_08


@pytest.mark.parametrize('backend', BYTES_BACKENDS)
def test_byte_decoder_decorator(backend):
    @backend.decoder
    @bpack.descriptor(baseunits=bpack.EBaseUnits.BYTES,
                      byteorder=bpack.EByteOrder.BE,
                      frozen=True)
    class Record:
        field_1: int = bpack.field(size=1, default=1)
        field_2: int = bpack.field(size=2, default=2)

    decoded_data = Record()
    encoded_data = bytes([0b00000001, 0b00000000, 0b00000010])
    record = Record.frombytes(encoded_data)

    assert record.field_1 == decoded_data.field_1
    assert record.field_2 == decoded_data.field_2


@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_unsupported_type(backend):
    class CustomType:
        pass

    with pytest.raises(TypeError):
        @backend.decoder
        @bpack.descriptor(baseunits=backend.Decoder.baseunits, frozen=True)
        class Record:                                                   # noqa
            field_1: CustomType = bpack.field(size=8)


@pytest.mark.parametrize('backend', BYTES_BACKENDS)
def test_byte_decoder_native_byteorder(backend):
    size = 4
    value = 1

    @backend.decoder
    @bpack.descriptor(byteorder=bpack.EByteOrder.NATIVE, frozen=True)
    class Record:
        field_1: int = bpack.field(size=size, default=value)

    data = value.to_bytes(size, sys.byteorder)
    assert Record.frombytes(data) == Record()


@pytest.mark.parametrize(
    'backend',
    [pytest.param(bpack_bs, id='bs',
                  marks=skipif(not bpack_bs, reason='not available'))])
def test_bit_decoder_native_byteorder(backend):
    size = 8
    value = 1

    @backend.decoder
    @bpack.descriptor(baseunits=bpack.EBaseUnits.BITS,
                      byteorder=bpack.EByteOrder.NATIVE,
                      frozen=True)
    class Record:
        field_1: int = bpack.field(size=8, default=1)

    data = value.to_bytes(size, sys.byteorder)
    assert Record.frombytes(data) == Record()


@pytest.mark.parametrize('backend', BITS_BACKENDS)
def test_bit_decoder_default_byteorder(backend):
    size = 8
    value = 1

    @backend.decoder
    @bpack.descriptor(baseunits=bpack.EBaseUnits.BITS,
                      byteorder=bpack.EByteOrder.DEFAULT,
                      frozen=True)
    class Record:
        field_1: int = bpack.field(size=8, default=1)

    # default byte order is big for bit descriptors
    data = value.to_bytes(size // 8, 'big')
    assert Record.frombytes(data) == Record()


@pytest.mark.parametrize('backend', BITS_BACKENDS)
def test_wrong_baseunits_bit(backend):
    with pytest.raises(ValueError):
        @backend.decoder
        @bpack.descriptor(baseunits=bpack.EBaseUnits.BYTES)
        class Record:                                                   # noqa
            field_1: int = bpack.field(size=8, default=1)


@pytest.mark.parametrize('backend', BYTES_BACKENDS)
def test_wrong_baseunits_byte(backend):
    with pytest.raises(ValueError):
        @backend.decoder
        @bpack.descriptor(baseunits=bpack.EBaseUnits.BITS)
        class Record:                                                   # noqa
            field_1: int = bpack.field(size=8, default=1)


@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_enum_decoding_bytes(backend):
    class EStrEnumType(enum.Enum):
        A = 'a'
        B = 'b'

    class EBytesEnumType(enum.Enum):
        A = b'a'
        B = b'b'

    class EIntEnumType(enum.Enum):
        A = 1
        B = 2

    class EFlagEnumType(enum.Enum):
        A = 1
        B = 2

    if backend.Decoder.baseunits is bpack.EBaseUnits.BYTES:
        bitorder = None
        ssize = 1
        isize = 1
        encoded_data = b''.join([
            EStrEnumType.A.value.encode('ascii'),
            EBytesEnumType.A.value,
            EIntEnumType.A.value.to_bytes(1, 'little', signed=False),
            EFlagEnumType.A.value.to_bytes(1, 'little', signed=False),
        ])
    else:
        bitorder = bpack.EBitOrder.MSB
        ssize = 8
        isize = 4
        encoded_data = b''.join([
            EStrEnumType.A.value.encode('ascii'),
            EBytesEnumType.A.value,
            bytes([0b00010001])
        ])

    @backend.decoder
    @bpack.descriptor(baseunits=backend.Decoder.baseunits, bitorder=bitorder)
    class Record:
        field_1: EStrEnumType = bpack.field(size=ssize, default=EStrEnumType.A)
        field_2: EBytesEnumType = bpack.field(size=ssize,
                                              default=EBytesEnumType.A)
        field_3: EIntEnumType = bpack.field(size=isize, default=EIntEnumType.A)
        field_4: EFlagEnumType = bpack.field(size=isize,
                                             default=EFlagEnumType.A)

    record = Record.frombytes(encoded_data)
    assert record == Record()


@pytest.mark.parametrize(
    'backend',
    [pytest.param(bpack.st, id='st'),
     pytest.param(bpack_bs, id='bs',
                  marks=skipif(not bpack_bs, reason='not available'))])
def test_sequence(backend):
    if backend.Decoder.baseunits is bpack.EBaseUnits.BYTES:
        bitorder = None
        size = 1
        repeat = 2
        encoded_data = bytes([3, 3, 4, 4])
    else:
        bitorder = bpack.EBitOrder.MSB
        size = 4
        repeat = 2
        encoded_data = bytes([0b00110011, 0b01000100])

    @backend.decoder
    @bpack.descriptor(baseunits=backend.Decoder.baseunits, bitorder=bitorder)
    class Record:
        field_1: List[int] = bpack.field(size=size, repeat=repeat)
        field_2: Sequence[int] = bpack.field(size=size, repeat=repeat)

    ref_record = Record([3, 3], (4, 4))
    record = Record.frombytes(encoded_data)
    assert record == ref_record

    assert type(record.field_1) is list
    assert type(record.field_2) is tuple

    for field, sequence_type in zip(bpack.fields(Record),
                                    (List[int], Sequence[int])):
        assert field.type == sequence_type


@pytest.mark.parametrize(
    'backend',
    [pytest.param(bpack.st, id='st'),
     pytest.param(bpack_bs, id='bs',
                  marks=skipif(not bpack_bs, reason='not available')),
     pytest.param(bpack_np, id='np',
                  marks=skipif(not bpack_np, reason='not available'))])
class TestNestedRecord:
    @staticmethod
    def get_encoded_data(baseunits):
        if baseunits is bpack.EBaseUnits.BYTES:
            encoded_data = bytes([
                0b00000000, 0b00000000, 0b00000000, 0b00000000,
                0b00000001, 0b00000000, 0b00000000, 0b00000000,
                0b00000010, 0b00000000, 0b00000000, 0b00000000,
                0b00000011, 0b00000000, 0b00000000, 0b00000000,
            ])

        else:  # baseunits is bpack.EBaseUnits.BITS:
            encoded_data = bytes([0b00000001, 0b00100011])

        return encoded_data

    def test_nested_record_decoders(self, backend):
        encoded_data = self.get_encoded_data(backend.Decoder.baseunits)

        @backend.decoder
        @bpack.descriptor(baseunits=backend.Decoder.baseunits)
        class Record:
            field_1: int = bpack.field(size=4, default=1)
            field_2: int = bpack.field(size=4, default=2)

        @backend.decoder
        @bpack.descriptor(baseunits=backend.Decoder.baseunits)
        class NestedRecord:
            field_1: int = bpack.field(size=4, default=0)
            field_2: Record = Record()
            field_3: int = bpack.field(size=4, default=3)

        assert NestedRecord.frombytes(encoded_data) == NestedRecord()

    def test_nested_record(self, backend):
        encoded_data = self.get_encoded_data(backend.Decoder.baseunits)

        @bpack.descriptor(baseunits=backend.Decoder.baseunits)
        class Record:
            field_1: int = bpack.field(size=4, default=1)
            field_2: int = bpack.field(size=4, default=2)

        @backend.decoder
        @bpack.descriptor(baseunits=backend.Decoder.baseunits)
        class NestedRecord:
            field_1: int = bpack.field(size=4, default=0)
            field_2: Record = Record()
            field_3: int = bpack.field(size=4, default=3)

        assert NestedRecord.frombytes(encoded_data) == NestedRecord()



"""Tests for codec utils."""

import pytest

import bpack
import bpack.st

from bpack.codecs import has_codec, get_codec, get_codec_type

try:
    import bpack.bs as bpack_bs
except ImportError:                                         # pragma: no cover
    bpack_bs = None


@pytest.mark.parametrize('backend',
                         [pytest.param(bpack.st, id='st'),
                          pytest.param(bpack_bs, id='bs',
                                       marks=pytest.mark.skipif(
                                           not bpack_bs,
                                           reason='not available'))])
def test_decoder_helpers(backend):
    @backend.decoder
    @bpack.descriptor(baseunits=backend.Decoder.baseunits)
    class Record:
        field_1: int = bpack.field(size=4, default=0)
        field_2: int = bpack.field(size=4, default=0)

    assert has_codec(Record)
    assert has_codec(Record, bpack.codecs.Decoder)
    assert has_codec(Record, backend.Decoder)
    assert get_codec_type(Record) is backend.Decoder
    assert isinstance(get_codec(Record), backend.Decoder)
    assert isinstance(get_codec(Record), bpack.codecs.Decoder)

    record = Record()
    assert has_codec(record)
    assert has_codec(record, bpack.codecs.Decoder)
    assert has_codec(record, backend.Decoder)
    assert get_codec_type(record) is backend.Decoder
    assert isinstance(get_codec(record), backend.Decoder)
    assert isinstance(get_codec(record), bpack.codecs.Decoder)
